package gameElements;

/**
 * Represents a card. It is comparable by the card rank.
 */
public class Card implements Comparable<Card>
{
    /**
     * The JACK rank.
     */
    public static final int JACK_RANK = 11;

    /**
     * The QUEEN rank.
     */
    public static final int QUEEN_RANK = 12;

    /**
     * The KING rank.
     */
    public static final int KING_RANK = 13;

    /**
     * The ACE rank.
     */
    public static final int ACE_RANK = 14;

    /**
     * The minimal card rank.
     */
    public static final int MINIMAL_CARD_RANK = 2;

    /**
     * The maximal card rank.
     */
    public static final int MAXIMAL_CARD_RANK = 14;

    /**
     * The hearts penalty.
     */
    public static final int HEARTS_PENALTY = 1;

    /**
     * The queen of spades penalty.
     */
    public static final int QUEEN_OF_SPADES_PENALTY = 13;

    /**
     * The suit of the card.
     */
    private final Suit suit;

    /**
     * The rank of the card.
     */
    private final int rank;

    /**
     * Constructs a new card from the given suit and the given rank.
     *
     * @param suit The given suit.
     * @param rank The given rank.
     */
    Card(Suit suit, int rank)
    {
        this.suit = suit;
        this.rank = rank;
    }

    /**
     * Gets the string representation of the rank of the card.
     */
    private String getRankAsString()
    {
        if (rank == JACK_RANK)
            return "jack";

        if (rank == QUEEN_RANK)
            return "queen";

        if (rank == KING_RANK)
            return "king";

        if (rank == ACE_RANK)
            return "ace";

        return Integer.toString(rank);
    }

    /**
     * Gets the penalty of the card.
     */
    public int getPenalty()
    {
        if (suit.equals(Suit.HEARTS))
            return HEARTS_PENALTY;

        if (suit.equals(Suit.SPADES) && rank == QUEEN_RANK)
            return QUEEN_OF_SPADES_PENALTY;

        return 0;
    }

    /**
     * Gets the suit of the card.
     */
    public Suit getSuit()
    {
        return suit;
    }

    /**
     * Gets the rank of the card.
     */
    public int getRank()
    {
        return rank;
    }

    /**
     * Compares the card to the given card by their rank.
     *
     * @param givenCard The card to be compared with.
     *
     * @return A negative number, if the card has less rank than the given one. Zero, if the cards have the same rank. A
     * positive number, if the card has higher rank than the given one.
     */
    @Override
    public int compareTo(Card givenCard)
    {
        return Integer.compare(rank, givenCard.getRank());
    }

    /**
     * Returns a string representation of the card.
     */
    @Override
    public String toString()
    {
        return getRankAsString() + "_of_" + suit.toString().toLowerCase();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode()
    {
        final int prime = 31;

        int result = 1;

        result = prime * result + rank;
        result = prime * result + ((suit == null) ? 0 : suit.hashCode());

        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;

        if (obj == null)
            return false;

        if (getClass() != obj.getClass())
            return false;

        Card other = (Card) obj;

        return rank == other.rank && suit == other.suit;
    }

    /**
     * The enumeration which represents the basically known card suits of a card deck.
     */
    public enum Suit
    {
        CLUBS, SPADES, DIAMONDS, HEARTS;

        /**
         * Gets an array of values of card suits in the wanted order. It replaces the implicit method values() of the
         * enumeration type, which is not recommended to use when we want a specific order of the enumeration elements.
         *
         * @return An array of the cards' suit values.
         */
        public static Suit[] getValues()
        {
            return new Suit[]{CLUBS, SPADES, DIAMONDS, HEARTS};
        }
    }
}