package gameElements;

import java.util.*;

/**
 * Represents a deck of cards. It is designed as a singleton. Provides methods for getting cards and their dealing.
 */
public class CardDeck
{
    /**
     * The instance of the card deck.
     */
    private static CardDeck instance;

    /**
     * The mapping of the card suits to the list of all cards of the given suit. It contains all the cards of deck.
     */
    private final Map<Card.Suit, List<Card>> cardMap;

    /**
     * The card list holding all cards of the deck.
     */
    private final List<Card> cardList;

    /**
     * Constructs a new card deck.
     */
    private CardDeck()
    {
        cardMap = new HashMap<>();
        cardList = new ArrayList<>();
        initializeCardMapAndList();
    }

    /**
     * Initializes the card map and the card list, i.e. creates the instances and fills their with cards.
     */
    private void initializeCardMapAndList()
    {
        for (Card.Suit suit : Card.Suit.getValues())
        {
            List<Card> cardsOfGivenSuit = new ArrayList<>();

            for (int i = Card.MINIMAL_CARD_RANK; i <= Card.MAXIMAL_CARD_RANK; i++)
            {
                Card card = new Card(suit, i);
                cardsOfGivenSuit.add(card);
                cardList.add(card);
            }

            cardMap.put(suit, cardsOfGivenSuit);
        }
    }

    /**
     * Gets a list of lists of cards, representing a list of hands. Each hand list contains the same number of cards (by
     * default 13). The size of the list equal to the number of players of a game (by default 4). The lists have
     * mutually different cards. The card distribution is pseudo-random (using Collections.shuffle()).
     *
     * @return The list of hand cards lists.
     */
    List<List<Card>> getDeal()
    {
        List<List<Card>> result = new ArrayList<>();

        Collections.shuffle(cardList);

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            List<Card> handCards = new ArrayList<>();

            final int tmp = Game.NUMBER_OF_CARDS_IN_HAND;

            for (int j = 0; j < tmp; j++)
                handCards.add(cardList.get(tmp * i + j));

            result.add(handCards);
        }

        return result;
    }

    /**
     * Gets a new (i.e. not a reference to the internal) list of all cards of the given suit.
     *
     * @param suit The given suit.
     *
     * @return A new list of all cards of the given suit.
     *
     * @throws NullPointerException If the suit is null.
     */
    public List<Card> getListOfCardsOfSuit(Card.Suit suit)
    {
        if (suit == null)
            throw new NullPointerException("The suit must not be null.");

        return new ArrayList<>(cardMap.get(suit));
    }

    /**
     * Gets a new (i.e. not a reference to the internal) list of all cards of the given suit.
     */
    public List<Card> getListOfAllCards()
    {
        return new ArrayList<>(cardList);
    }

    /**
     * Gets the queen of spades.
     */
    public Card getQueenOfSpades()
    {
        return cardMap.get(Card.Suit.SPADES).get(Card.QUEEN_RANK - Card.MINIMAL_CARD_RANK);
    }

    /**
     * Gets the two of clubs.
     */
    public Card getTwoOfClubs()
    {
        return cardMap.get(Card.Suit.CLUBS).get(2 - Card.MINIMAL_CARD_RANK);
    }

    /**
     * Gets the instance of the card deck.
     */
    public static CardDeck getInstance()
    {
        if (instance == null)
            instance = new CardDeck();

        return instance;
    }
}