package gameElements;

import infranstructure.GameException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Represents a game of given players (by default 4), i.e. the set of rounds which are being played until there is a
 * player with a minimal ending score (by default 100).
 */
public class Game
{
    /**
     * The number of players in a game.
     */
    public static final int NUMBER_OF_PLAYERS = 4;

    /**
     * The number of cards in each player's hand. Also the number of tricks of each round.
     */
    public static final int NUMBER_OF_CARDS_IN_HAND = 13;

    /**
     * The number of cards which are changed at the beginning of each non-fourth round.
     */
    public static final int NUMBER_OF_CARDS_TO_CHANGE = 3;

    /**
     * The minimal score, which if some player has, the game ends.
     */
    public static final int MINIMAL_ENDING_SCORE = 100;

    /**
     * The maximal penalty, which a player could get in a round.
     */
    public static final int MAXIMAL_ROUND_PENALTY = 26;

    /**
     * The list of players in the game.
     */
    private final List<Player> players;

    /**
     * The current round of the game.
     */
    private Round currentRound;

    /**
     * The serial number of the current round, numbered from 1.
     */
    private int currentRoundNumber = 0;

    /**
     * Construct a new game of the given list of players.
     *
     * @param players The list of players of the game.
     *
     * @throws NullPointerException     If the list of players is null.
     * @throws IllegalArgumentException If the number of players isn't equal to the wanted number of players (4 by
     *                                  default).
     */
    public Game(List<Player> players)
    {
        if (players == null)
            throw new NullPointerException("The players of a game must not be null.");

        if (players.size() != Game.NUMBER_OF_PLAYERS)
            throw new IllegalArgumentException("The number of players of a game must be " + Game.NUMBER_OF_PLAYERS + ".");

        this.players = players;

        for (Player player : players)
            player.setCurrentGame(this);
    }

    /**
     * Prepares a new round to be played, i.e. replaces the current round with a new one.
     */
    public void prepareNewRound()
    {
        currentRoundNumber++;
        currentRound = new Round(this, currentRoundNumber);
    }

    /**
     * Indicates whether the game has ended, i.e. there is a player who already has the minimal ending score.
     *
     * @return true if it has ended, false otherwise
     */
    public boolean hasGameEnded()
    {
        for (Player player : players)
            if (player.getCurrentGameScore() >= MINIMAL_ENDING_SCORE)
                return true;

        return false;
    }

    /**
     * Gets the index of the winner of the game, or -1, if the game ended with a draw.
     *
     * @return The index of the winner, if there is a winner, or -1, if there is a draw.
     *
     * @throws GameException If the game hasn't ended yet.
     */
    public int getIndexOfWinner()
    {
        if (!hasGameEnded())
            throw new GameException("There's no winner yet");

        List<Player> allPlayers = new ArrayList<>(players);

        Collections.sort(allPlayers);

        Player playerWithTheLeastScore = allPlayers.get(0);

        if (playerWithTheLeastScore.compareTo(allPlayers.get(1)) == 0)
            return -1;

        return players.indexOf(playerWithTheLeastScore);
    }

    /**
     * Gets the list of players in the game.
     *
     * @return The list of players.
     */
    public List<Player> getPlayers()
    {
        return players;
    }

    /**
     * Gets the current round of the game.
     *
     * @return The current round.
     */
    public Round getCurrentRound()
    {
        return currentRound;
    }

    /**
     * Gets the current round's serial number (numbered from 1).
     *
     * @return The current round's serial number.
     */
    public int getCurrentRoundNumber()
    {
        return currentRoundNumber;
    }
}