package gameElements;

import gameElements.utilities.CardUtilities;

import java.util.ArrayList;
import java.util.List;

/**
 * Represent a player's hand, i.e. the set of cards which player gets at the beginning of a round.
 */
public class Hand
{
    /**
     * The list of cards of the hand.
     */
    private final List<Card> cardsOfHand;

    /**
     * Construct a new hand from the given list of cards and sorts the cards.
     *
     * @param cardsOfHand The given list of cards.
     **/
    Hand(List<Card> cardsOfHand)
    {
        this.cardsOfHand = cardsOfHand;
        CardUtilities.sortListOfCards(this.cardsOfHand);
    }

    /**
     * Plays the card of the hand, i.e. removes it from the hand.
     *
     * @param card The card to be played.
     */
    void playCard(Card card)
    {
        cardsOfHand.remove(card);
    }

    /**
     * Passes the list of cards to another player (i.e. removes them from the hand), according to the rules of passing
     * of the current round. Defaultly there should be 3 cards to be passed.
     *
     * @param toBePassed The list of cards to be passed.
     */
    void passCards(List<Card> toBePassed)
    {
        cardsOfHand.removeIf(toBePassed::contains);
    }

    /**
     * Adds cards, which someone has passed me, to the hand.
     *
     * @param receivedCards The list of cards from the pass to be added to the hand.
     */
    void addCardsWhichSomeoneHasPassedMe(List<Card> receivedCards)
    {
        cardsOfHand.addAll(receivedCards);
        CardUtilities.sortListOfCards(cardsOfHand);
    }

    /**
     * Returns if the hand (i.e. the list of hand cards) contains the given card.
     *
     * @param card The given card.
     *
     * @return true, if the hand contains the card, false otherwise
     */
    public boolean contains(Card card)
    {
        return cardsOfHand.contains(card);
    }

    /**
     * Gets the new (i.e. not a reference to the internal) list of cards of the hand.
     *
     * @return A new list of the hand cards.
     */
    public List<Card> getHandCardsList()
    {
        return new ArrayList<>(cardsOfHand);
    }
}