package gameElements;

import gameElements.suggestors.MoveSuggester;
import gameElements.suggestors.PassSuggester;
import gameElements.utilities.GameUtilities;
import infranstructure.CardException;
import infranstructure.GameException;
import infranstructure.RoundException;

import java.util.ArrayList;
import java.util.List;

/**
 * Represents a player of a game. It is comparable by the current game score.
 */
public class Player implements Comparable<Player>
{
    /**
     * The player's name.
     */
    private final String name;

    /**
     * The current player's game.
     */
    private Game currentGame;

    /**
     * The current player's hand.
     */
    private Hand currentHand;

    /**
     * The current player's game score.
     */
    private int currentGameScore;

    /**
     * The current player's round score.
     */
    private int currentRoundScore;

    /**
     * The number of moonshots in the current game.
     */
    private int numberOfCurrentGameMoonshots;

    /**
     * The list of the round scores of the current game.
     */
    private List<Integer> currentGameRoundScores;

    /**
     * Construct a new player with the given name.
     *
     * @param name The name of the player.
     */
    public Player(String name)
    {
        this.name = name;
        currentGameRoundScores = new ArrayList<>();
    }

    /**
     * Gets the list of currently playable cards.
     *
     * @return The list of currently playable cards.
     *
     * @throws NullPointerException If the player hasn't been assigned to a game.
     * @throws GameException        If the game hasn't already prepared any round.
     * @throws CardException        If the cards haven't been dealt yet.
     */
    public List<Card> getCurrentlyPlayableCards()
    {
        if (currentGame == null)
            throw new NullPointerException("The player hasn't been assigned to a game.");

        if (currentGame.getCurrentRound() == null)
            throw new GameException("The game hasn't already prepared any round.");

        if (currentHand == null)
            throw new CardException("The cards haven't been dealt yet.");

        return GameUtilities.getPlayableCards(currentHand.getHandCardsList(), currentGame.getCurrentRound());
    }

    /**
     * Gets a move suggestion at the current state of the current game.
     *
     * @return The card that move suggester suggests to be played.
     *
     * @throws NullPointerException If the player hasn't been assigned to a game.
     * @throws GameException        If the game hasn't already prepared any round.
     * @throws CardException        If the cards haven't been dealt yet.
     * @throws RoundException       If the first trick hasn't been prepared yet.
     */
    public Card getMoveSuggestion()
    {
        if (currentGame == null)
            throw new NullPointerException("The player hasn't been assigned to a game.");

        if (currentGame.getCurrentRound() == null)
            throw new GameException("The game hasn't already prepared any round.");

        if (currentHand == null)
            throw new CardException("The cards haven't been dealt yet.");

        if (currentGame.getCurrentRound().getCurrentTrick() == null)
            throw new RoundException("The first trick hasn't been prepared yet.");

        return new MoveSuggester(this).suggest();
    }

    /**
     * Gets a pass suggestion at the current state of the current game.
     *
     * @return The list of cards that suggester suggests to be passed.
     *
     * @throws NullPointerException If the player hasn't been assigned to a game.
     * @throws GameException        If the game hasn't already prepared any round.
     * @throws CardException        If the cards haven't been dealt yet.
     * @throws RoundException       If the passing has been done or the round is a no passing round.
     */
    public List<Card> getPassSuggestion()
    {
        if (currentGame == null)
            throw new NullPointerException("The player hasn't been assigned to a game.");

        if (currentGame.getCurrentRound() == null)
            throw new GameException("The game hasn't already prepared any round.");

        if (currentHand == null)
            throw new CardException("The cards haven't been dealt yet.");

        if (currentGame.getCurrentRound().getCurrentTrick() != null)
            throw new RoundException("The passing has been done or the round is a no passing round.");

        return new PassSuggester(this).suggest();
    }

    /**
     * Gets the score of the round with the given serial number.
     *
     * @param serialNumber The serial number of the round.
     *
     * @return The score of the round.
     *
     * @throws IndexOutOfBoundsException If the round with the given serial number hasn't ended yet.
     */
    public int getScoreOfRound(int serialNumber)
    {
        int index = serialNumber - 1;

        if (index < 0 || index >= currentGameRoundScores.size())
            throw new IndexOutOfBoundsException("Cannot ask for a round score of the round with serial number: " + serialNumber);

        return currentGameRoundScores.get(serialNumber - 1);
    }

    /**
     * Updates current round score according to the given trick, i.e. gets the total trick penalty and adds it to the
     * current rounds score.
     *
     * @param trick The trick according to which will be the score updated.
     */
    void updateRoundScore(Trick trick)
    {
        currentRoundScore += trick.getTotalPenalty();
    }

    /**
     * Updates the current game score, i.e. adds the current round to the current game score and then sets the current
     * round score to 0.
     */
    void updateGameScore()
    {
        currentGameScore += currentRoundScore;
        currentGameRoundScores.add(currentRoundScore);
        currentRoundScore = 0;
    }

    /**
     * Records the moonshot i.e. increases the number of current game moonshots by one.
     */
    void recordMoonshot()
    {
        numberOfCurrentGameMoonshots++;
    }

    /**
     * Sets the current round score.
     *
     * @param currentRoundScore The score to be set.
     */
    void setCurrentRoundScore(int currentRoundScore)
    {
        this.currentRoundScore = currentRoundScore;
    }

    /**
     * Sets the current hand.
     *
     * @param hand The hand to be set.
     */
    void setCurrentHand(Hand hand)
    {
        currentHand = hand;
    }

    /**
     * Sets the current game and assigns all the variables connected with the current game to it's default value.
     *
     * @param currentGame The game to be set.
     */
    void setCurrentGame(Game currentGame)
    {
        this.currentGame = currentGame;
        currentHand = null;
        currentGameScore = 0;
        currentRoundScore = 0;
        numberOfCurrentGameMoonshots = 0;
        currentGameRoundScores = new ArrayList<>();
    }

    /**
     * Gets the current game.
     *
     * @return The current game.
     */
    public Game getCurrentGame()
    {
        return currentGame;
    }

    /**
     * Gets the current hand.
     *
     * @return The current hand.
     */
    public Hand getCurrentHand()
    {
        return currentHand;
    }

    /**
     * Gets the current round score.
     *
     * @return The current round score.
     */
    int getCurrentRoundScore()
    {
        return currentRoundScore;
    }

    /**
     * Gets the current game score.
     *
     * @return The current game score.
     */
    public int getCurrentGameScore()
    {
        return currentGameScore;
    }

    /**
     * Gets the number of moonshots in the current game.
     *
     * @return The number of moonshots in the current game.
     */
    public int getNumberOfMoonshots()
    {
        return numberOfCurrentGameMoonshots;
    }

    /**
     * Returns a string representation of the player, i.e. their name.
     *
     * @return The name of the player.
     */
    @Override
    public String toString()
    {
        return name;
    }

    /**
     * Compares the player to the given player by their current game score.
     *
     * @return A negative number, if the player has less current game score than the given one. Zero, if the player have
     * the same current game score. A positive number, if the card has higher current game score than the given one.
     */
    @Override
    public int compareTo(Player givenPlayer)
    {
        return Integer.compare(getCurrentGameScore(), givenPlayer.getCurrentGameScore());
    }
}