package gameElements;

import gameElements.utilities.SuitWatcher;
import infranstructure.CardException;
import infranstructure.RoundException;

import java.util.ArrayList;
import java.util.List;

/**
 * Represents a round of the given game. The class provides the main methods of the game background, playCard and
 * passCards.
 */
public class Round
{
    /**
     * The game which the round belongs to.
     */
    private final Game game;

    /**
     * The integer representation of the type of passing in this round. It's the serial round number modulo 4. <br> <br>
     * 0 - no pass round <br> 1 - each player passes right <br> 2 - each player passes left <br> 3 - each player passes
     * across
     */
    private final int typeOfPassing;

    /**
     * Represents a bijection of the players passing such that the player with the index i receives the cards from the
     * player with the index passingPairs[i]. It's null in the no pass round. It is initialized by the method
     * initializePassingPairs.
     */
    private int[] passingPairs;

    /**
     * The number of players who currently passed their cards.
     */
    private int numberOfPasses;

    /**
     * List of lists of cards to be passed. The cards are being passed only after each player has chosen the cards for
     * being passed. Up until then, the players passes are stored in this list.
     */
    private List<List<Card>> listOfCardsToBePassed;

    /**
     * Indicates whether the passes have been evaluated, i.e. the passed cards have been changed between the players.
     */
    private boolean passesEvaluated;

    /**
     * The current trick in the round.
     */
    private Trick currentTrick;

    /**
     * The serial number of the current trick.
     */
    private int currentTrickNumber;

    /**
     * The index of the current trick starter.
     */
    private int indexOfCurrentTrickStarter;

    /**
     * The index of the current player on turn.
     */
    private int indexOfPlayerOnTurn;

    /**
     * Indicates whether hearts have been broken.
     */
    private boolean heartsBroken;

    /**
     * The list of cards which have been already played in the round.
     */
    private final List<Card> playedCards;

    /**
     * The suit watcher of the current round.
     */
    private final SuitWatcher suitWatcher;

    /**
     * Indicates whether the round has ended.
     */
    private boolean hasEnded;

    /**
     * Construct a new round of the given game with the given serial number.
     *
     * @param game         The given game.
     * @param serialNumber The given serial number.
     */
    Round(Game game, int serialNumber)
    {
        this.game = game;
        typeOfPassing = serialNumber % Game.NUMBER_OF_PLAYERS;
        initializePassingPairs();
        playedCards = new ArrayList<>();
        suitWatcher = new SuitWatcher(this);
        dealCards();
        prepareCardPassing();
    }

    /**
     * Deals the cards to the players.
     */
    private void dealCards()
    {
        List<List<Card>> handLists = CardDeck.getInstance().getDeal();

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
            getPlayers().get(i).setCurrentHand(new Hand(handLists.get(i)));
    }

    /**
     * Prepares card passing. If this is no pass round, it prepares a round start.
     */
    private void prepareCardPassing()
    {
        if (isThisNoChangeRound())
        {
            prepareRoundStart();
            return;
        }

        listOfCardsToBePassed = new ArrayList<>();

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
            listOfCardsToBePassed.add(null);
    }

    /**
     * Pass the given cards of the player with the given index.
     *
     * @param indexOfPasser   The index of passer.
     * @param cardsToBePassed The cards to be passed.
     *
     * @throws IndexOutOfBoundsException If the indexOfPasser is out of bounds, i.e. it's less than 0 and greater or
     *                                   equal than the number of players of a game.
     * @throws NullPointerException      If the cardsToBePassed are null.
     * @throws CardException             If the number of cards to be passed isn't equal to the required number (3 by
     *                                   default).
     * @throws RoundException            If this is no passing round or the passing has been finished.
     */
    public void passCards(int indexOfPasser, List<Card> cardsToBePassed)
    {
        if (indexOfPasser < 0 || indexOfPasser >= Game.NUMBER_OF_PLAYERS)
            throw new IndexOutOfBoundsException("There is no player with index " + indexOfPasser);

        if (cardsToBePassed == null)
            throw new NullPointerException("Card to be changed must not be null");

        if (cardsToBePassed.size() != Game.NUMBER_OF_CARDS_TO_CHANGE)
            throw new CardException("Cannot change more than " + Game.NUMBER_OF_CARDS_TO_CHANGE + " cards.");

        if (isThisNoChangeRound())
            throw new RoundException("In this round there is no card changing.");

        if (numberOfPasses >= Game.NUMBER_OF_PLAYERS)
            throw new RoundException("Changing has been finished.");

        listOfCardsToBePassed.set(indexOfPasser, cardsToBePassed);
        numberOfPasses++;

        if (numberOfPasses == Game.NUMBER_OF_PLAYERS)
            evaluatePassing();
    }

    /**
     * Evaluates the passing, i.e. changes cards between players according to the passing rules of the round. Then
     * prepares a round start.
     */
    private void evaluatePassing()
    {
        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            Hand currentHand = getPlayers().get(i).getCurrentHand();
            currentHand.passCards(listOfCardsToBePassed.get(i));
            currentHand.addCardsWhichSomeoneHasPassedMe(listOfCardsToBePassed.get(passingPairs[i]));
        }

        prepareRoundStart();
    }

    /**
     * Prepares round start, i.e. finds the round starter and calls the method prepareNewTrick.
     */
    private void prepareRoundStart()
    {
        passesEvaluated = true;
        indexOfCurrentTrickStarter = getRoundStarter();
        prepareNewTrick();
    }

    /**
     * Gets the index of the round starter, i.e. the index of player who holds in his hand the two of clubs.
     *
     * @return The index of the round starter.
     */
    private int getRoundStarter()
    {
        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
            if (getPlayers().get(i).getCurrentHand().contains(CardDeck.getInstance().getTwoOfClubs()))
                return i;

        return -1;
    }

    /**
     * Prepares a new trick.
     */
    public void prepareNewTrick()
    {
        if (hasEnded)
            return;

        currentTrick = new Trick(indexOfCurrentTrickStarter);
        indexOfPlayerOnTurn = indexOfCurrentTrickStarter;
        currentTrickNumber++;
    }

    /**
     * Plays the next move of the current trick. The player who plays the card is automatically the player on turn.
     *
     * @param cardToBePlayed The card to be played.
     *
     * @throws NullPointerException If the cardToBePlayed is null
     * @throws RoundException       If the passes haven't been evaluated yet.
     * @throws CardException        If the card is not currently playable for the player on turn.
     */
    public void playNextMove(Card cardToBePlayed)
    {
        if (cardToBePlayed == null)
            throw new NullPointerException("cardToBePlayed must not be null");

        if (!passesEvaluated)
            throw new RoundException("Cannot make a move until changes are evaluated");

        Player player = getPlayers().get(indexOfPlayerOnTurn);

        if (!player.getCurrentlyPlayableCards().contains(cardToBePlayed))
            throw new CardException("The card " + cardToBePlayed + " is not currently playable");

        currentTrick.addCard(cardToBePlayed);
        playedCards.add(cardToBePlayed);
        suitWatcher.updateMap();

        if (cardToBePlayed.getSuit() == Card.Suit.HEARTS)
            heartsBroken = true;

        getPlayers().get(indexOfPlayerOnTurn).getCurrentHand().playCard(cardToBePlayed);

        if (currentTrick.isFull())
        {
            evaluateCurrentTrick();
            return;
        }

        indexOfPlayerOnTurn = getIndexOfTheNextPlayerOnTurn();
    }

    /**
     * Evaluates the current trick, i.e. finds a winner, updates their score and prepares a new trick (unless the round
     * has ended).
     */
    private void evaluateCurrentTrick()
    {
        int indexOfWinner = currentTrick.getIndexOfWinner();
        Player winner = getPlayers().get(indexOfWinner);
        winner.updateRoundScore(currentTrick);

        if (currentTrickNumber != Game.NUMBER_OF_CARDS_IN_HAND)
        {
            indexOfCurrentTrickStarter = indexOfWinner;
            return;
        }

        updatePlayersScore();
        hasEnded = true;
    }

    /**
     * Update players score after the end of the round.
     */
    private void updatePlayersScore()
    {
        Player moonShooter = getPlayerWhoShotAMoon();

        if (moonShooter != null)
        {
            moonShooter.recordMoonshot();
            setRoundScoresAfterMoonshot(moonShooter);
        }

        getPlayers().forEach(Player::updateGameScore);
    }

    /**
     * Find a player who shot the moon.
     *
     * @return A player who shot the moon, or null, if there is no such a player.
     */
    private Player getPlayerWhoShotAMoon()
    {
        for (Player player : getPlayers())
            if (player.getCurrentRoundScore() == Game.MAXIMAL_ROUND_PENALTY)
                return player;

        return null;
    }

    /**
     * Sets the round scores to the players if there is a player who shot a moon.
     *
     * @param moonShooter The moon shooter.
     */
    private void setRoundScoresAfterMoonshot(Player moonShooter)
    {
        for (Player player : getPlayers())
        {
            if (player.equals(moonShooter))
                player.setCurrentRoundScore(0);
            else
                player.setCurrentRoundScore(Game.MAXIMAL_ROUND_PENALTY);
        }
    }

    /**
     * Gets the index of the next player on turn.
     *
     * @return The index of the next player on turn.
     */
    private int getIndexOfTheNextPlayerOnTurn()
    {
        return indexOfPlayerOnTurn == Game.NUMBER_OF_PLAYERS - 1 ? 0 : indexOfPlayerOnTurn + 1;
    }

    /**
     * Initializes the field passingPairs, according to the type of passing of the round.
     */
    private void initializePassingPairs()
    {
        if (typeOfPassing == 1)
            passingPairs = new int[]{3, 0, 1, 2};

        if (typeOfPassing == 2)
            passingPairs = new int[]{1, 2, 3, 0};

        if (typeOfPassing == 3)
            passingPairs = new int[]{2, 3, 0, 1};
    }

    /**
     * Gets the way of passing of the round as a string.
     *
     * @return The string representation of the way of passing.
     */
    public String getTheWayOfPassingAsString()
    {
        if (typeOfPassing == 1)
            return "Pass left";

        if (typeOfPassing == 2)
            return "Pass right";

        if (typeOfPassing == 3)
            return "Pass across";

        return null;
    }

    /**
     * Gets received cards of the player with the given index.
     *
     * @param index The given index.
     *
     * @return Received cards of the player with the index.
     */
    public List<Card> getReceivedCardsOfPlayer(int index)
    {
        return listOfCardsToBePassed.get(passingPairs[index]);
    }

    /**
     * Returns if this is no passing round.
     *
     * @return true, if this is no passing round, false otherwise
     */
    public boolean isThisNoChangeRound()
    {
        return typeOfPassing == 0;
    }

    /**
     * Gets the current trick.
     *
     * @return The current trick.
     */
    public Trick getCurrentTrick()
    {
        return currentTrick;
    }

    /**
     * Gets the index of the current trick starter.
     *
     * @return The index of the current trick starter.
     */
    public int getIndexOfCurrentTrickStarter()
    {
        return indexOfCurrentTrickStarter;
    }

    /**
     * Returns if the current trick is first.
     *
     * @return true, if the current trick is first, false otherwise
     */
    public boolean isTheCurrentTrickFirst()
    {
        return currentTrickNumber == 1;
    }

    /**
     * Returns if the hearts are already broken.
     *
     * @return true, if the hearts are already broken, false otherwise
     */
    public boolean areHeartsBroken()
    {
        return heartsBroken;
    }

    /**
     * Returns if the round has ended.
     *
     * @return true, if the round has ended, false otherwise
     */
    public boolean hasRoundEnded()
    {
        return hasEnded;
    }

    public int getIndexOfLastTrickWinner()
    {
        if (!hasEnded)
            return indexOfCurrentTrickStarter;

        return currentTrick.getIndexOfWinner();
    }

    /**
     * Gets the index of player, who is currently on turn.
     *
     * @return The index of player, who is currently on turn.
     */
    public int getIndexOfPlayerOnTurn()
    {
        return indexOfPlayerOnTurn;
    }

    /**
     * Gets the player, who is currently on turn.
     *
     * @return The player, who is currently on turn.
     */
    public Player getPlayerOnTurn()
    {
        return getPlayers().get(indexOfPlayerOnTurn);
    }

    /**
     * Returns if the player with the given index is on turn.
     *
     * @param indexOfPlayer The given player's index.
     *
     * @return true, if the player is on turn, false otherwise
     */
    public boolean isPlayerOnTurn(int indexOfPlayer)
    {
        return indexOfPlayerOnTurn == indexOfPlayer;
    }

    /**
     * Gets a new (i.e. not a reference to the internal) list of cards currently played in the round.
     *
     * @return A new list of the played cards.
     */
    public List<Card> getPlayedCardsList()
    {
        return new ArrayList<>(playedCards);
    }

    /**
     * Gets the round's suit watcher.
     *
     * @return The round's suit watcher.
     */
    public SuitWatcher getSuitWatcher()
    {
        return suitWatcher;
    }

    /**
     * Gets the list of players of the current game.
     *
     * @return The list of players of the current game.
     */
    public List<Player> getPlayers()
    {
        return game.getPlayers();
    }

    public int[] getPassingPairs()
    {
        return passingPairs;
    }
}