package gameElements;

import gameElements.utilities.CardUtilities;
import infranstructure.CardException;

import java.util.ArrayList;
import java.util.List;

/**
 * Represents a trick in a round, i.e. the set of cards played by each player in turn, during the play of a hand.
 */
public class Trick
{
    /**
     * The cards currently played in the trick.
     */
    private final List<Card> playedCards;

    /**
     * The index of the trick starter.
     */
    private final int indexOfStarter;

    /**
     * The index of the trick winner.
     */
    private int indexOfWinner;

    /**
     * Indicates whether the trick has already been evaluated, i.e. it has been decided who its winner is.
     */
    private boolean evaluated;

    /**
     * Constructs an empty trick with the given trick starter.
     *
     * @param indexOfStarter index of the trick starter.
     */
    Trick(int indexOfStarter)
    {
        playedCards = new ArrayList<>();
        this.indexOfStarter = indexOfStarter;
    }

    /**
     * Adds the card to the trick.
     *
     * @param card The card to be added.
     */
    void addCard(Card card)
    {
        playedCards.add(card);

        if (isFull())
            evaluate();
    }

    /**
     * Evaluates the trick, i.e. decides who is the winner and assigns the variable indexOfWinner.
     */
    private void evaluate()
    {
        Card winningCard = CardUtilities.getHighestCardWithSuitAsFirst(playedCards);
        int indexOfWinningCard = playedCards.indexOf(winningCard);
        indexOfWinner = (indexOfWinningCard + indexOfStarter) % Game.NUMBER_OF_PLAYERS;
        evaluated = true;
    }

    /**
     * Returns whether the trick is empty, i.e. there is no played card yet.
     *
     * @return true, if the trick is empty, else otherwise.
     */
    public boolean isEmpty()
    {
        return playedCards.isEmpty();
    }

    /**
     * Returns whether the trick is full, i.e. it has the maximal number of cards inside, 4 by default.
     *
     * @return true, if the trick is full, false otherwise.
     */
    public boolean isFull()
    {
        return playedCards.size() == Game.NUMBER_OF_PLAYERS;
    }

    /**
     * Gets the total penalty for the current state of the trick.
     *
     * @return The total penalty for the trick.
     */
    int getTotalPenalty()
    {
        return CardUtilities.getTotalPenaltyOf(playedCards);
    }

    /**
     * Gets the index of the winner of the trick.
     *
     * @return The index of the winner of the trick.
     *
     * @throws CardException The trick hasn't been evaluated yet.
     */
    int getIndexOfWinner()
    {
        if (!evaluated)
            throw new CardException("The trick hasn't been evaluated yet.");

        return indexOfWinner;
    }

    /**
     * Gets a new (i.e. not a reference to the internal) list of cards currently played in the trick.
     *
     * @return A new list of the played cards.
     */
    public List<Card> getPlayedCardsList()
    {
        return new ArrayList<>(playedCards);
    }
}