package gameElements.suggestors;

import gameElements.Card;
import gameElements.CardDeck;
import gameElements.Player;
import gameElements.utilities.CardAnalyzer;
import gameElements.utilities.CardUtilities;
import gameElements.utilities.EquivalencyMap;

import java.util.List;

/**
 * A helper class, which can suggest a move to a player at the given time of the game. The suggestion is made by
 * heuristic algorithms.
 */
public class MoveSuggester
{
    /**
     * The cards that have been played in the current trick.
     */
    private final List<Card> trickCards;

    /**
     * The player's cards, stored in the equivalency map.
     */
    private final EquivalencyMap cards;

    /**
     * The card analyzer for the player.
     */
    private final CardAnalyzer analyzer;

    /**
     * The queen of spades.
     */
    private final Card queenOfSpades;

    /**
     * The suit of the first card played in the current trick, or null, if no card has been played.
     */
    private final Card.Suit suitOfFirst;

    /**
     * Indicates if the current player is the trick starter, i.e. if the current trick is empty.
     */
    private final boolean isTrickStarter;

    /**
     * Indicates if the queen has been already played in the in the current round.
     */
    private final boolean isQueenOut;

    /**
     * Indicates if the player has the queen.
     */
    private final boolean hasQueen;

    /**
     * Indicates if the player has high spades, i.e. King of Spades, or Ace of Spades.
     */
    private final boolean hasHighSpades;

    /**
     * Indicates if the current trick is the first in game.
     */
    private final boolean isThisFirstTrick;

    /**
     * Indicates if the hears bas been broken already in the current round.
     */
    private final boolean areHeartsBroken;

    /**
     * Constructs a move suggester for the given player.
     */
    public MoveSuggester(Player player)
    {
        List<Card> handCards = player.getCurrentHand().getHandCardsList();
        List<Card> playableCards = player.getCurrentlyPlayableCards();
        List<Card> playedCards = player.getCurrentGame().getCurrentRound().getPlayedCardsList();

        trickCards = player.getCurrentGame().getCurrentRound().getCurrentTrick().getPlayedCardsList();
        cards = new EquivalencyMap(playableCards, playedCards, trickCards);
        analyzer = new CardAnalyzer(player);
        queenOfSpades = CardDeck.getInstance().getQueenOfSpades();
        isTrickStarter = trickCards.isEmpty();
        isQueenOut = playedCards.contains(queenOfSpades);
        hasQueen = handCards.contains(queenOfSpades);
        hasHighSpades = cards.isThereCardOfTheSameSuitHigherThan(queenOfSpades);
        isThisFirstTrick = player.getCurrentGame().getCurrentRound().isTheCurrentTrickFirst();
        areHeartsBroken = player.getCurrentGame().getCurrentRound().areHeartsBroken();
        suitOfFirst = isTrickStarter ? null : trickCards.get(0).getSuit();
    }

    /**
     * Suggests a card that the player could play.
     */
    public Card suggest()
    {
        // if there is basically no option, we'll return a random card
        if (cards.isThereOnlyOneClassOfEquivalence())
            return cards.getRandomCard();

        if (isTrickStarter)
            return playAsTrickStarter();

        return playAsNonTrickStarter();
    }

    /**
     * Suggests a card if I am a trick starter
     */
    private Card playAsTrickStarter()
    {
        // if the queen is out, we play as we don't want to win it. It's not always the best strategy,
        // because we may stay left with high cards
        if (isQueenOut)
            return playTrickStartNotToWinIt();

        if (hasQueen)
            return playTrickStartIfIHaveQueen();

        // if we can lure the queen (see doc. of the method canLureQueen()), we'll do it and the best
        // strategy for that is to play as highest spades less than queen as possible.
        if (canLureQueen())
            return cards.getRandomHighestCardOfGivenSuitLessThan(queenOfSpades);

        // otherwise we just don't want to start the next trick. Again, at this point the logic could be
        // improved a lot, because I believe this works in at most 2/3 of cases.
        return playTrickStartNotToWinIt();
    }

    /**
     * Suggests a card if I am a trick starter and have the queen of spades.
     */
    private Card playTrickStartIfIHaveQueen()
    {
        // if there is someone, who will surely take the trick if there is the queen in it,
        // we'll play the queen
        if (analyzer.isThereAPlayerWhoWillHaveToPlayHigherThan(queenOfSpades))
            return queenOfSpades;

        // otherwise we'll try to get rid of the cards of some suit (distinct from spades)
        return playCardToGetRidOfSuitDistinctFromSpadesAsStarterWithQueen();
    }

    /**
     * Determines whether it's safe to lure the queen of not. This means, to start playing spades less than the queen
     * until the player who has the queen has to play it.
     */
    private boolean canLureQueen()
    {
        final int mySpades = cards.getNumberOfCardsOfTheSuitAsGivenLessThanGiven(queenOfSpades);

        // if we have no spades lower than the queen, we surely won't succeed
        if (mySpades == 0)
            return false;

        // we assume that someone has the queen so there is at least one spade left in the game
        final int spadesLeft = cards.getNumberOfCardsOfGivenSuitLeftInGame(Card.Suit.SPADES);

        // if we have more other spades than there are spades left in game, we'll surely succeed.
        if (mySpades >= spadesLeft)
            return true;

        // if we don't have high spades, we have nothing to lose. Otherwise we can't be sure that
        // we'll lure the queen before we have only high spades
        return !hasHighSpades;
    }

    /**
     * Suggests a card to get rid of a suit distinct from spades (we'll need them to handle spade-attacks). We assume we
     * have the queen and we start the trick.
     */
    private Card playCardToGetRidOfSuitDistinctFromSpadesAsStarterWithQueen()
    {
        List<Card.Suit> availableSuits = getSuitsThatIHaveAndSomeoneCouldHaveThemToo();
        availableSuits.remove(Card.Suit.SPADES);

        // if there is a diamond or club to be get rid of and there is a chance that someone
        // else would take it then we'll play it
        if (availableSuits.contains(Card.Suit.DIAMONDS) || availableSuits.contains(Card.Suit.CLUBS))
        {
            availableSuits.remove(Card.Suit.HEARTS);
            List<Card> cardsToPlay = cards.getListOfCardsOfSuitWhichIHaveTheLeast(availableSuits);
            return cards.getRandomHighestCardFromList(cardsToPlay);
        }

        // now we know that even if we have diamonds or clubs, no one else has them. Let's assume
        // we don't have only spades and hearts aren't broken. We might try to break them. But we surely
        // can't play hearts (because we have at least the queen of spades). So we must have a diamond or a club
        // and it doesn't matter which one will be played (because no one else has them)
        if (!areHeartsBroken && !cards.areThereOnlyCardsOfTheSuit(Card.Suit.SPADES))
            return cards.getRandomCardDistinctFromSpades();

        // at this point, we'll never play clubs or diamonds. No one else has them. If hearts are broken, then
        // we wouldn't achieve anything by playing them. Otherwise we have only spades. First we check if there
        // is a change to play hearts and don't lost the trick

        List<Card> cardsToPlay = cards.getNonLostCardsOfGivenSuit(Card.Suit.HEARTS);

        if (!cardsToPlay.isEmpty())
            return cards.getRandomLowestCardOfSuit(Card.Suit.HEARTS);

        // if there are only hearts with we can only lost the trick, then we
        // won't play them. So we're gonna check spades

        cardsToPlay = cards.getNonLostCardsOfGivenSuit(Card.Suit.SPADES);

        // if there is no option to play, we are basically lost
        if (cardsToPlay.size() == 0)
            return cards.getRandomCard();

        // if there is the only option, we'll just play it and hope that someone will take it
        if (cardsToPlay.size() == 1)
            return queenOfSpades;

        // otherwise we'll just try to play as lowest spade as possible
        return cards.getRandomLowestSpadeDistinctFromQueen();
    }

    /**
     * Suggests a card if we are a trick starter and don't want to win the trick.
     */
    private Card playTrickStartNotToWinIt()
    {
        // if we don't have any not-lost card, we are basically done
        if (cards.areNonLostCardsEmpty())
            return cards.getRandomCard();

        // otherwise we'll just try to play the lowest card from the cards of the suit that has the most cards
        // left in the game, if we take into account only not-lost cards
        return cards.getRandomLowestCardOfSuitWithMostCardsLeftInGameFromNotLostCards();
    }

    /**
     * Suggests a card if we aren't a trick starter.
     */
    private Card playAsNonTrickStarter()
    {
        if (isThisFirstTrick)
            return playTheFirstTrick();

        if (cards.isThereCardOfTheGivenSuit(suitOfFirst))
            return playAsNonTrickStarterWithSuitOfFirst();

        return playAsNonTrickStarterWithoutSuitOfFirst();
    }

    /**
     * Suggest a card for the first trick, if we are not a trick starter.
     */
    private Card playTheFirstTrick()
    {
        // it's very common to play as highest club as we have, if we have one
        if (cards.isThereCardOfTheGivenSuit(Card.Suit.CLUBS))
            return cards.getRandomHighestCardOfGivenSuit(Card.Suit.CLUBS);

        // if we have the queen, we want to get rid of a suit distinct from spades
        if (hasQueen)
            return playCardToGetRidOfSuit(true);

        // if we don't have the queen, but have high spades, we'll get rid of the highest one
        if (hasHighSpades)
            return cards.getRandomHighestCardOfGivenSuit(Card.Suit.SPADES);

        // in every other case we'll just try to get rid of a suit
        return playCardToGetRidOfSuit(false);
    }

    /**
     * Suggest a card for at least the second trick, if we are not a trick starter and don't have the suit of the first
     * played card.
     */
    private Card playAsNonTrickStarterWithoutSuitOfFirst()
    {
        // if we have the queen, we surely can play it
        if (hasQueen)
            return queenOfSpades;

        // if we don't have the queen, but high spades and they are dangerous (i.e. the queen isn't out yet),
        // we'll get rid of the highest one
        if (hasHighSpades && !isQueenOut)
            return cards.getRandomHighestCardOfGivenSuit(Card.Suit.SPADES);

        // if spades aren't interesting and we have hearts, we'll just play the highest one
        if (cards.isThereCardOfTheGivenSuit(Card.Suit.HEARTS))
            return cards.getRandomHighestCardOfGivenSuit(Card.Suit.HEARTS);

        // in every other case we''ll just to try get rid of any suit
        return playCardToGetRidOfSuit(false);
    }

    /**
     * Suggest a card, if we are not a trick start and we have the suit of the first played card.
     */
    private Card playAsNonTrickStarterWithSuitOfFirst()
    {
        // if there is the queen in the trick, we just don't want to take it (i.e. we want to dump the trick)
        if (trickCards.contains(queenOfSpades))
            return playDumpingCurrentTrick();

        // if we have to play spades and we have the queen, well resolve this case individually
        if (Card.Suit.SPADES.equals(suitOfFirst) && hasQueen)
            return playSpadesAsNonTrickStarterIfIHaveToPlaySpadesAndHaveQueen();

        // if there is a heart in the trick, we'll just dump it (it's always the most save solution,
        // but not the best one)
        if (CardUtilities.isThereCardOfGivenSuit(trickCards, Card.Suit.HEARTS))
            return playDumpingCurrentTrick();

        // if it's safe to play the highest card, we'll do it
        if (isSafeToPlayHighest())
            return cards.getRandomHighestCardOfGivenSuit(suitOfFirst);

        // in every other case we'll try to dump the trick
        return playDumpingCurrentTrick();
    }

    /**
     * Suggests a spade card, if we have to play spades, have the queen and we're not a trick starter.
     */
    private Card playSpadesAsNonTrickStarterIfIHaveToPlaySpadesAndHaveQueen()
    {
        // if some high spades has been already played, we could play the queen
        if (CardUtilities.isThereCardOfTheSameSuitAndHigherRank(trickCards, queenOfSpades))
            return queenOfSpades;

        // otherwise we'll just not to be take the next turn
        return cards.getRandomLowestSpadeDistinctFromQueen();
    }

    /**
     * Determines if it's safe to play the highest card, if we are not a trick starter, we have the suit of the first
     * played card and there are no hearts in the trick
     */
    private boolean isSafeToPlayHighest()
    {
        Card highest = getHighestCardThatMustBePlayedInCurrentTrick();

        // if someone has to play a higher card than our's highest, then it's safe to play the highest
        if (cards.getRandomHighestCardOfGivenSuit(suitOfFirst).compareTo(highest) < 0)
            return true;

        boolean playersAfterMeDoNotHaveHearts = analyzer.allPlayersAfterMeDoNotHaveSuit(Card.Suit.HEARTS);

        // if the queen doesn't concern us, then we're sure if the players after us don't have hearts
        if (hasQueen || isQueenOut)
            return playersAfterMeDoNotHaveHearts;

        // otherwise the queen might be dangerous, if there is a player after us who has spades
        if (!analyzer.allPlayersAfterMeDoNotHaveSuit(Card.Suit.SPADES))
            return false;

        // otherwise we're sure if the players after us don't have hearts
        return playersAfterMeDoNotHaveHearts;
    }

    /**
     * Suggest a card such that we don't want to take the trick, if we are not a trick starter.
     */
    private Card playDumpingCurrentTrick()
    {
        Card highestInTrick = getHighestCardThatMustBePlayedInCurrentTrick();
        Card possiblePlay = cards.getRandomHighestCardOfGivenSuitLessThan(highestInTrick);

        // if we can play the card which is less than the highest card that must be played, we'll do it
        if (possiblePlay != null)
            return possiblePlay;

        // if there is a chance that someone after us would play a higher card, we'll try to play
        // as lowest card as possible
        if (analyzer.isThereAPlayerWhoCouldPlayHigherThan(highestInTrick))
            return cards.getRandomLowestCardOfSuit(suitOfFirst);

        // in every other case we can't dump the trick, so we'll play the highest possible card
        return cards.getRandomHighestCardOfGivenSuit(suitOfFirst);
    }

    /**
     * Suggests a card, if we want to get rid of some suit, with an option whether take spades into account or not.
     */
    private Card playCardToGetRidOfSuit(boolean exceptSpades)
    {
        List<Card.Suit> availableSuits = getSuitsThatIHaveAndSomeoneCouldHaveThemToo();

        if (exceptSpades)
            availableSuits.remove(Card.Suit.SPADES);

        // if no one could have the suits that we have, we are screwed
        if (availableSuits.isEmpty())
            return cards.getRandomCard();

        // otherwise we'll just take a random card of the suit that we have the least
        return CardUtilities.getRandomCard(cards.getListOfCardsOfSuitWhichIHaveTheLeast(availableSuits));
    }

    /**
     * Finds the highest card that must be played in the current trick by some player after us.
     */
    private Card getHighestCardThatMustBePlayedInCurrentTrick()
    {
        Card highestInTrick = CardUtilities.getHighestCardWithSuitAsFirst(trickCards);
        Card possibleHighest = analyzer.getHighestCardOfGivenSuitThatSomeoneMustPlay(suitOfFirst);

        // if possibleHighest exists and it's higher than the currently highest, then the new highest is the
        // possible highest
        if (possibleHighest != null && possibleHighest.compareTo(highestInTrick) > 0)
            highestInTrick = possibleHighest;

        return highestInTrick;
    }

    /**
     * From the suits that we have chooses the ones which someone else after us could have as well.
     */
    private List<Card.Suit> getSuitsThatIHaveAndSomeoneCouldHaveThemToo()
    {
        List<Card.Suit> availableSuits = cards.getListOfSuitsThatIHave();
        List<Card.Suit> suitsThatSomeoneCouldHave = analyzer.getListOfSuitsThatSomeoneCouldHave();
        availableSuits.retainAll(suitsThatSomeoneCouldHave);
        return availableSuits;
    }
}