package gameElements.suggestors;

import gameElements.Card;
import gameElements.CardDeck;
import gameElements.Player;
import gameElements.utilities.CardUtilities;

import java.util.ArrayList;
import java.util.List;

/**
 * A helper class, which can suggest handCards to be passed to a player at the given round of the game. The suggestion
 * is made by heuristic algorithms.
 */
public class PassSuggester
{
    /**
     * The list of cards in the player's hand.
     */
    private final List<Card> handCards;

    /**
     * The queen of spades.
     */
    private final Card queenOfSpades;

    /**
     * The number of spades in the hand.
     */
    private final int numberOfSpades;

    /**
     * The number of high spades in the hand, i.e. King of Spades, or Ace of Spades.
     */
    private final int numberOfHighSpades;

    /**
     * Indicates if the player has the queen.
     */
    private final boolean hasQueen;

    /**
     * Indicates if the player has high spades, i.e. King of Spades, or Ace of Spades.
     */
    private final boolean hasHighSpades;

    /**
     * Constructs a pass suggester for the given player.
     */
    public PassSuggester(Player player)
    {
        handCards = player.getCurrentHand().getHandCardsList();
        queenOfSpades = CardDeck.getInstance().getQueenOfSpades();
        numberOfSpades = CardUtilities.getNumberOfCardsOfGivenSuit(handCards, Card.Suit.SPADES);
        numberOfHighSpades = CardUtilities.getNumberOfCardsOfGivenSuitHigherThan(handCards, queenOfSpades);
        hasQueen = handCards.contains(queenOfSpades);
        hasHighSpades = CardUtilities.isThereCardOfTheSameSuitAndHigherRank(handCards, queenOfSpades);
    }

    /**
     * Suggests a list of cards that player could pass.
     */
    public List<Card> suggest()
    {
        if (hasQueen)
            return passIfIHaveQueen();

        if (hasHighSpades)
            return passWithoutQueenWithHighSpades();

        // if we don't have the queen or high spades, we'll just pass the highest cards
        return CardUtilities.getHighestCards(handCards, 3, true);
    }

    /**
     * Suggests cards to be passed if we have the queen.
     */
    private List<Card> passIfIHaveQueen()
    {
        // if we don't have enough spades, we'll pass the queen and the highest spades, and the remaining high
        // cards
        if (numberOfSpades <= 3)
        {
            List<Card> result = new ArrayList<>();
            result.add(queenOfSpades);
            result.addAll(CardUtilities.getHighSpades(handCards));
            List<Card> notChosenCards = CardUtilities.getAllCardsExceptCardsFrom(handCards, result);
            result.addAll(CardUtilities.getHighestCards(notChosenCards, 3 - result.size(), true));
            return result;
        }

        // otherwise we'll just pass high cards, if it's possible, non-spades
        return getHighestCardWithPriorityForNonSpades(new ArrayList<>());
    }

    /**
     * Suggests cards to be passed if we don't have the queen, but we have high spades.
     */
    private List<Card> passWithoutQueenWithHighSpades()
    {
        List<Card> result = new ArrayList<>();

        // if we have at most three spades, with 2 high ones, we'll pass the high ones
        if (numberOfHighSpades == 2 && numberOfSpades <= 3)
            result.addAll(CardUtilities.getHighSpades(handCards));

        // if we have at most two spades, with one a high one, we'll pass the high one
        if (numberOfHighSpades == 1 && numberOfSpades <= 2)
            result.addAll(CardUtilities.getHighSpades(handCards));

        result.addAll(getHighestCardWithPriorityForNonSpades(result));

        return result;
    }

    /**
     * Gets list of highest cards from the hand, that will be passed together with the list of currently passed cards.
     * The size of the returned list will be 3 - size of the current list.
     */
    private List<Card> getHighestCardWithPriorityForNonSpades(List<Card> current)
    {
        List<Card> result = new ArrayList<>();

        List<Card> notPassedCards = CardUtilities.getAllCardsExceptCardsFrom(handCards, current);
        int numberOfCards = 3 - current.size();
        int numberOfNonSpades = CardUtilities.getAllCardsWithoutGivenSuit(notPassedCards, Card.Suit.SPADES).size();
        int numberOfNonSpadesToBePassed = Math.min(numberOfCards, numberOfNonSpades);

        // we'll pass highest non-spades
        result.addAll(CardUtilities.getHighestCards(handCards, numberOfNonSpadesToBePassed, false));
        // and if there are at most 2 of them, we'll pass also highest spades lower than queen
        result.addAll(CardUtilities.getHighestSpadesLowerThanQueen(handCards, numberOfCards - result.size()));

        return result;
    }
}