package gameElements.utilities;

import gameElements.Card;
import gameElements.CardDeck;
import gameElements.Player;

import java.util.*;

/**
 * A helper class, which allows a given player to predict cards of other player, according to their's own cards and the
 * current state of the game.
 */
public class CardAnalyzer
{
    /**
     * The mapping of players to the cards, which they could have, which is represented as a map of card suits to the
     * list of cards of the particular suit.
     */
    private final Map<Player, SortedMap<Card.Suit, List<Card>>> map;

    /**
     * The suit watcher of the current round.
     */
    private final SuitWatcher suitWatcher;

    /**
     * The reference to the list of all cards in the card deck.
     */
    private final List<Card> allCards;

    /**
     * The list of card that have been already played in the round.
     */
    private final List<Card> playedCards;

    /**
     * The list of card that the player currently has in their hand.
     */
    private final List<Card> handCards;

    /**
     * The list of all players of the game.
     */
    private final List<Player> allPlayers;

    /**
     * The list of cards that have been already played in the current trick.
     */
    private final List<Card> currentTrickCards;

    /**
     * The player, which uses the analyzer.
     */
    private final Player player;

    /**
     * The index of the current trick starter.
     */
    private final int indexOfTrickStarter;

    /**
     * Constructs a new analyzer for a given player.
     */
    public CardAnalyzer(Player player)
    {
        this.player = player;
        suitWatcher = player.getCurrentGame().getCurrentRound().getSuitWatcher();
        allCards = CardDeck.getInstance().getListOfAllCards();
        playedCards = player.getCurrentGame().getCurrentRound().getPlayedCardsList();
        handCards = player.getCurrentHand().getHandCardsList();
        allPlayers = player.getCurrentGame().getPlayers();
        indexOfTrickStarter = player.getCurrentGame().getCurrentRound().getIndexOfCurrentTrickStarter();
        currentTrickCards = player.getCurrentGame().getCurrentRound().getCurrentTrick().getPlayedCardsList();
        map = new HashMap<>();
        initializeMap();
    }

    /**
     * Initializes the main map.
     */
    private void initializeMap()
    {
        for (Player player : allPlayers)
        {
            // we don't have to estimate our cards
            if (player.equals(this.player))
                continue;

            map.put(player, new TreeMap<>());

            for (Card.Suit suit : Card.Suit.getValues())
                map.get(player).put(suit, new ArrayList<>());
        }

        fillMap();
    }

    /**
     * Fills the main map.
     */
    private void fillMap()
    {
        // we'll go through all cards
        for (Card card : allCards)
        {
            // if we have the card or it's been played, no one could have it
            if (handCards.contains(card) || playedCards.contains(card))
                continue;

            // otherwise we go through the all players in the map (we're sure we're not there)
            for (Player player : map.keySet())
            {
                Card.Suit suit = card.getSuit();

                // if the given player could have the suit (according to the suit watcher), then we
                // can add to card to their's map
                if (suitWatcher.couldPlayerHaveSuit(player, suit))
                    map.get(player).get(suit).add(card);
            }
        }
    }

    /**
     * Returns whether we can be sure that all players after me don't have a given suit.
     */
    public boolean allPlayersAfterMeDoNotHaveSuit(Card.Suit suit)
    {
        List<Player> playersAfterMe = GameUtilities
                .getPlayersWhoDidNotPlayTrick(allPlayers, indexOfTrickStarter, currentTrickCards.size());
        // we are obviously included, because we haven't played the trick yet
        playersAfterMe.remove(player);

        for (Player player : playersAfterMe)
            if (!map.get(player).get(suit).isEmpty())
                return false;

        return true;
    }

    /**
     * Finds the highest card of a given suit that some player after us must play (i.e. they have no option).
     */
    public List<Card.Suit> getListOfSuitsThatSomeoneCouldHave()
    {
        List<Player> playersAfterMe = GameUtilities
                .getPlayersWhoDidNotPlayTrick(allPlayers, indexOfTrickStarter, currentTrickCards.size());
        // we are obviously included, because we haven't played the trick yet
        playersAfterMe.remove(player);

        Set<Card.Suit> result = new HashSet<>();

        for (Player player : playersAfterMe)
            for (Card.Suit suit : Card.Suit.getValues())
                if (!map.get(player).get(suit).isEmpty())
                    result.add(suit);

        return new ArrayList<>(result);
    }

    /**
     * Finds the highest card of the given suit that some player after us must play in the current trick. If there is no
     * such card (or rather we can't be sure that there is), it will return null.
     */
    public Card getHighestCardOfGivenSuitThatSomeoneMustPlay(Card.Suit suit)
    {
        List<Player> playersAfterMe = GameUtilities
                .getPlayersWhoDidNotPlayTrick(allPlayers, indexOfTrickStarter, currentTrickCards.size());
        // we are obviously included, because we haven't played the trick yet
        playersAfterMe.remove(player);

        List<Card> cardsThatSomeoneMustPlay = new ArrayList<>();

        // we'll go through the all players except us
        for (Player player : allPlayers)
        {
            if (player.equals(this.player))
                continue;

            List<Card> cardsThatPlayerCouldHave = map.get(player).get(suit);

            // if the current player doesn't have the suit, then we skip the player
            if (cardsThatPlayerCouldHave.isEmpty())
                continue;

            // otherwise the current player could play a card of the given suit. And if the player isn't
            // after us, we can't be sure that the player don't have all the remaining cards of the given suit,
            // therefore we must return null.
            if (!playersAfterMe.contains(player))
                return null;

            cardsThatSomeoneMustPlay.add(Collections.min(cardsThatPlayerCouldHave));
        }

        // this happens when we are the only player who has cards of the given suit
        if (cardsThatSomeoneMustPlay.isEmpty())
            return null;

        // at this point, we are sure that the players different from us, who could have cards of the given suit,
        // are the ones who'll play after us, so someone will have to play a card
        return Collections.max(cardsThatSomeoneMustPlay);
    }

    /**
     * Determines whether there is a player after us who could play a card higher than a given one.
     */
    public boolean isThereAPlayerWhoCouldPlayHigherThan(Card card)
    {
        List<Player> playersAfterMe = GameUtilities
                .getPlayersWhoDidNotPlayTrick(allPlayers, indexOfTrickStarter, currentTrickCards.size());
        // we are obviously included, because we haven't played the trick yet
        playersAfterMe.remove(player);

        for (Player player : playersAfterMe)
        {
            List<Card> cards = map.get(player).get(card.getSuit());

            if (CardUtilities.isThereCardOfTheSameSuitAndHigherRank(cards, card))
                return true;
        }

        return false;
    }

    /**
     * Determines whether there is a player after us who must play a card higher than a given one.
     */
    public boolean isThereAPlayerWhoWillHaveToPlayHigherThan(Card card)
    {
        Card higher = getHighestCardOfGivenSuitThatSomeoneMustPlay(card.getSuit());
        return higher != null && higher.compareTo(card) > 0;
    }
}