package gameElements.utilities;

import gameElements.Card;
import gameElements.CardDeck;

import java.util.*;
import java.util.stream.Collectors;

/**
 * Provides static utilities for cards.
 */
public class CardUtilities
{
    /**
     * From the card list creates a mapping of card suits to the list of cards of the given suit.
     */
    public static SortedMap<Card.Suit, List<Card>> getCardMapFromList(List<Card> cardList)
    {
        SortedMap<Card.Suit, List<Card>> result = new TreeMap<>();

        for (Card card : cardList)
        {
            Card.Suit suit = card.getSuit();

            if (!result.containsKey(suit))
                result.put(suit, new ArrayList<>());

            result.get(suit).add(card);
        }

        return result;
    }

    /**
     * Sorts a given list of cards such that it first groups cards by suits (in the order given in the declaration of
     * {@link Card.Suit}, and then in each group according to the {@link Card#compareTo(Card)}.
     */
    public static void sortListOfCards(List<Card> listOfCards)
    {
        List<Card> temp = new ArrayList<>();

        SortedMap<Card.Suit, List<Card>> cardMap = getCardMapFromList(listOfCards);

        for (List<Card> list : cardMap.values())
        {
            Collections.sort(list);
            temp.addAll(list);
        }

        ListIterator<Card> it = listOfCards.listIterator();

        for (Card card : temp)
        {
            it.next();
            it.set(card);
        }
    }

    /**
     * From the given list of cards selects the ones with the given suit.
     */
    public static List<Card> getListOfCardsOfGivenSuit(List<Card> cardList, Card.Suit suit)
    {
        return cardList.stream().filter(card -> card.getSuit().equals(suit)).collect(Collectors.toList());
    }

    /**
     * From a given list of cards selects the ones with zero penalty.
     */
    public static List<Card> getCardsWithZeroPenalty(List<Card> cardList)
    {
        return cardList.stream().filter(card -> card.getPenalty() == 0).collect(Collectors.toList());
    }

    /**
     * From the given list of cards selects that doesn't have the given suit.
     */
    public static List<Card> getAllCardsWithoutGivenSuit(List<Card> cardList, Card.Suit givenSuit)
    {
        List<Card> result = new ArrayList<>();

        for (Card.Suit suit : Card.Suit.getValues())
            if (!suit.equals(givenSuit))
                result.addAll(getListOfCardsOfGivenSuit(cardList, suit));

        return result;
    }

    /**
     * Determines whether a given list contains only hearts.
     */
    public static boolean hasHeartsOnly(List<Card> cardList)
    {
        return getListOfCardsOfGivenSuit(cardList, Card.Suit.HEARTS).size() == cardList.size();
    }

    /**
     * From a given card list selects a random card.
     */
    public static Card getRandomCard(List<Card> cardList)
    {
        return cardList.get((int) (Math.random() * cardList.size()));
    }

    /**
     * Determines whether a given list contains a card of a given suit.
     */
    public static boolean isThereCardOfGivenSuit(List<Card> cardList, Card.Suit suit)
    {
        return !getListOfCardsOfGivenSuit(cardList, suit).isEmpty();
    }

    /**
     * Determines whether a given list contains a card of the same suit as a given one, with a higher rank.
     */
    public static boolean isThereCardOfTheSameSuitAndHigherRank(List<Card> cardList, Card card)
    {
        List<Card> cardsOfGivenSuit = getListOfCardsOfGivenSuit(cardList, card.getSuit());
        return !cardsOfGivenSuit.isEmpty() && Collections.max(cardsOfGivenSuit).compareTo(card) > 0;
    }

    /**
     * Determines whether a given list contains a card of the same suit as a given one, with a lower rank.
     */
    public static boolean isThereCardOfTheSameSuitOfLowerRank(List<Card> cardList, Card card)
    {
        List<Card> cardsOfGivenSuit = getListOfCardsOfGivenSuit(cardList, card.getSuit());
        return !cardsOfGivenSuit.isEmpty() && Collections.min(cardsOfGivenSuit).compareTo(card) < 0;
    }

    /**
     * Gets all card suits that are contained in a given card list.
     */
    public static List<Card.Suit> getListOfCardSuits(List<Card> cardList)
    {
        return new ArrayList<>(cardList.stream().map(Card::getSuit).collect(Collectors.toSet()));
    }

    /**
     * Determines whether a given list contains all possible cards of a given suit.
     */
    public static boolean areThereAllPossibleCardsOfGivenSuit(List<Card> cardList, Card.Suit suit)
    {
        List<Card> list = CardDeck.getInstance().getListOfCardsOfSuit(suit);

        for (Card card : list)
            if (!cardList.contains(card))
                return false;

        return true;
    }

    /**
     * From a given list of cards selects the highest card with the same suit as the first card in the list.
     */
    public static Card getHighestCardWithSuitAsFirst(List<Card> cardList)
    {
        return Collections.max(getListOfCardsOfGivenSuit(cardList, cardList.get(0).getSuit()));
    }

    /**
     * From a given list of cards excludes all cards that are contained in another given list of cards to be excluded.
     */
    public static List<Card> getAllCardsExceptCardsFrom(List<Card> cardList, List<Card> cardsToBeExcluded)
    {
        return cardList.stream().filter(card -> !cardsToBeExcluded.contains(card)).collect(Collectors.toList());
    }

    /**
     * Gets the number of cards of a given suit in a given card list.
     */
    public static int getNumberOfCardsOfGivenSuit(List<Card> cardList, Card.Suit suit)
    {
        return getListOfCardsOfGivenSuit(cardList, suit).size();
    }

    /**
     * Gets the number of cards in a given card list, higher than a given card with the same suit.
     */
    public static int getNumberOfCardsOfGivenSuitHigherThan(List<Card> cardList, Card card)
    {
        int result = 0;

        for (Card cardFromList : cardList)
            if (cardFromList.getSuit().equals(card.getSuit()) && cardFromList.compareTo(card) > 0)
                result++;

        return result;
    }

    /**
     * From a given list of cards selects all high spades.
     */
    public static List<Card> getHighSpades(List<Card> cards)
    {
        return cards.stream().filter(card -> card.getSuit().equals(Card.Suit.SPADES) && card
                .compareTo(CardDeck.getInstance().getQueenOfSpades()) > 0).collect(Collectors.toList());
    }

    /**
     * From a given list of cards selects a given number of the highest spades lower than the queen.
     */
    public static List<Card> getHighestSpadesLowerThanQueen(List<Card> cards, int numberOfWantedCards)
    {
        List<Card> result = new ArrayList<>();

        if (numberOfWantedCards == 0)
            return result;

        List<Card> spades = getListOfCardsOfGivenSuit(cards, Card.Suit.SPADES);
        Collections.sort(spades);
        Collections.reverse(spades);

        for (int i = 0; i < cards.size(); i++)
        {
            Card spade = spades.get(i);

            if (spade.getRank() >= Card.QUEEN_RANK)
                continue;

            result.add(spade);

            if (result.size() == numberOfWantedCards)
                break;
        }

        return result;
    }

    /**
     * From a given list of cards selects a given number of the highest cards, with an option whether to take into
     * account spades or not.
     */
    public static List<Card> getHighestCards(List<Card> cards, int numberOfWantedCards, boolean includeSpades)
    {
        List<Card> result = new ArrayList<>();

        if (numberOfWantedCards == 0)
            return result;

        List<Card> copyOfCards = new ArrayList<>(cards);
        Collections.sort(copyOfCards);
        Collections.reverse(copyOfCards);

        for (int i = 0; i < cards.size(); i++)
        {
            Card card = copyOfCards.get(i);

            if (card.getSuit().equals(Card.Suit.SPADES) && !includeSpades)
                continue;

            result.add(card);

            if (result.size() == numberOfWantedCards)
                break;
        }

        return result;
    }

    /**
     * Gets the total penalty of the cards in the given list.
     */
    public static int getTotalPenaltyOf(List<Card> cardList)
    {
        int result = 0;

        for (Card card : cardList)
            result += card.getPenalty();

        return result;
    }
}