package gameElements.utilities;

import gameElements.Card;
import gameElements.CardDeck;

import java.util.*;
import java.util.stream.Collectors;

/**
 * A helper class, which sorts all cards into equivalency classes. For a given hand we say that two cards are equivalent
 * if and only they have the same suit and it doesn't matter which one is played - it would have the same effect anyway.
 * For example: If we have the clubs 2,3,5,6,7,J,Q,A, then the equivalency classes are: {2,3}, {5,6,7}, {J,Q}, {A}. So
 * we can also say that two cards are equivalent if and only if every card we have every card between them in the hand.
 * There is one special case. The queen of spades is obviously a different card from the others, so it makes sense to
 * define that the queen of spades is not equivalent to any card. For example: If we have the spades 4,7,8,J,Q,K,A, then
 * the equivalency classes are {4}, {7,8}, {J}, {Q}, {K,A}.
 */
public class EquivalencyMap
{
    /**
     * The mapping of the card suits to the cards of the suit represented as a mapping of the card ranks to the list of
     * cards equivalent to the card with that rank.
     */
    private final SortedMap<Card.Suit, SortedMap<Integer, List<Card>>> map;

    /**
     * The list of playable cards in the player's hand.
     */
    private final List<Card> playableCards;

    /**
     * The list of cards that have been currently played, except the ones that have been played in the current trick.
     */
    private final List<Card> playedCardsWithoutTrickCards;

    /**
     * The list of cards that have been currently played.
     */
    private final List<Card> playedCards;

    /**
     * The list of cards for which there is a player who has a higher card of the same suit.
     */
    private List<Card> nonLostCards;

    /**
     * Constructs an equivalency map.
     *
     * @param playableCards Cards, that are currently playable.
     * @param playedCards   Cards, that have been already played.
     * @param trickCards    Cards, that are in the current trick.
     */
    public EquivalencyMap(List<Card> playableCards, List<Card> playedCards, List<Card> trickCards)
    {
        map = new TreeMap<>();
        this.playableCards = playableCards;
        this.playedCardsWithoutTrickCards = CardUtilities.getAllCardsExceptCardsFrom(playedCards, trickCards);
        this.playedCards = playedCards;
        initializeMap();
        findNonLostCards();
    }

    /**
     * Initializes {@link #map}.
     */
    private void initializeMap()
    {
        for (Card.Suit suit : Card.Suit.getValues())
        {
            List<Card> handCardsOfGivenSuit = CardUtilities.getListOfCardsOfGivenSuit(playableCards, suit);

            if (handCardsOfGivenSuit.isEmpty())
                continue;

            List<Card> playedCardsOfGivenSuit = CardUtilities.getListOfCardsOfGivenSuit(playedCardsWithoutTrickCards, suit);
            map.put(suit, getMapOfEquivalencyClassesForOneSuit(handCardsOfGivenSuit, playedCardsOfGivenSuit));
        }
    }

    /**
     * Gets an equivalency class mapping from the given card list, with the given list of cards that have been played.
     * We assume that all the cards in both list have the same suit.
     */
    private SortedMap<Integer, List<Card>> getMapOfEquivalencyClassesForOneSuit(List<Card> cardList, List<Card> playedCards)
    {
        SortedMap<Integer, List<Card>> result = new TreeMap<>();

        Collections.sort(cardList);

        int currentIndex = 0;
        Card currentCard = cardList.get(0);
        result.put(currentCard.getRank(), new ArrayList<>());
        result.get(currentCard.getRank()).add(currentCard);

        for (int i = 1; i < cardList.size(); i++)
        {
            Card card = cardList.get(i);

            if (areTwoCardsEquivalent(i, currentIndex, cardList, playedCards))
            {
                result.get(currentCard.getRank()).add(card);
                continue;
            }

            result.put(card.getRank(), new ArrayList<>());
            result.get(card.getRank()).add(card);
            currentCard = card;
            currentIndex = i;
        }

        return result;
    }

    /**
     * Finds out whether two cards, given by its indices in the hands cards list, are equivalent in the current state of
     * the game. We assume that all the cards in both lists have the same suit.
     */
    private boolean areTwoCardsEquivalent(int indexOfFirst, int indexOfSecond, List<Card> handCards, List<Card> playedCards)
    {
        Card lower = handCards.get(indexOfFirst);
        Card higher = handCards.get(indexOfSecond);

        Card queen_of_spades = CardDeck.getInstance().getQueenOfSpades();

        // The queen of spades is never equivalent to any of card.
        if (lower.equals(queen_of_spades) || higher.equals(queen_of_spades))
            return false;

        // We make that the lower card has really a lower rank.
        if (lower.compareTo(higher) > 0)
        {
            Card tmp = lower;
            lower = higher;
            higher = tmp;
        }

        // Now the low card is equivalent to the higher one if and only if every card with a rank between them two has
        // been played or we hold the card in the hand.
        for (int i = lower.getRank() + 1; i < higher.getRank(); i++)
            if (!isThereCardWithRank(i, playedCards) && !isThereCardWithRank(i, handCards))
                return false;

        return true;
    }

    /**
     * Finds out whether the given card list contains a card with the given rank.
     */
    private boolean isThereCardWithRank(int rank, List<Card> cardList)
    {
        for (Card card : cardList)
            if (card.getRank() == rank)
                return true;

        return false;
    }

    /**
     * Finds {@link #nonLostCards}.
     */
    private void findNonLostCards()
    {
        nonLostCards = playableCards.stream().filter(this::canBeBeaten).collect(Collectors.toList());
    }

    /**
     * Finds out whether the given card can be beaten by some another card left in the game.
     */
    private boolean canBeBeaten(Card card)
    {
        List<Card> cardsOfGivenSuit = CardDeck.getInstance().getListOfCardsOfSuit(card.getSuit());

        // A card can be beaten if and only if from the cards of the same suit we can choose a card with a higher
        // rank that hasn't been played yet and we don't hold it.
        for (Card cardOfGivenSuit : cardsOfGivenSuit)
            if (cardOfGivenSuit.compareTo(card) > 0)
                if (!playedCards.contains(cardOfGivenSuit) && !playableCards.contains(cardOfGivenSuit))
                    return true;

        return false;
    }

    /**
     * Finds out whether it doesn't matter which card would be played, because all of them are equivalent to each
     * other.
     */
    public boolean isThereOnlyOneClassOfEquivalence()
    {
        return map.keySet().size() == 1 && map.get(map.firstKey()).keySet().size() == 1;
    }

    /**
     * Finds out whether the playable cards contains only cards of one suit.
     */
    public boolean areThereOnlyCardsOfTheSuit(Card.Suit suit)
    {
        return map.containsKey(suit) && map.keySet().size() == 1;
    }

    /**
     * Finds out whether the playable cards contains a card with the same suit as the given one, and a higher rank.
     */
    public boolean isThereCardOfTheSameSuitHigherThan(Card card)
    {
        return CardUtilities.isThereCardOfTheSameSuitAndHigherRank(playableCards, card);
    }

    /**
     * Finds out whether the playable cards contains a card with the same suit as the given one, and a lower rank.
     */
    public boolean isThereCardOfTheSameSuitLowerThan(Card card)
    {
        return CardUtilities.isThereCardOfTheSameSuitOfLowerRank(playableCards, card);
    }

    /**
     * Finds out whether the playable cards contains a card of the given suit.
     */
    public boolean isThereCardOfTheGivenSuit(Card.Suit suit)
    {
        return map.containsKey(suit);
    }

    /**
     * Find out whether there is such a playable card that can
     */
    public boolean areNonLostCardsEmpty()
    {
        return nonLostCards.isEmpty();
    }

    /**
     * Gets a random card from the playable cards.
     */
    public Card getRandomCard()
    {
        return CardUtilities.getRandomCard(playableCards);
    }

    /**
     * Gets a random card of the given suit from the playable cards.
     */
    public Card getRandomHighestCardOfGivenSuit(Card.Suit suit)
    {
        return CardUtilities.getRandomCard(map.get(suit).get(map.get(suit).lastKey()));
    }

    /**
     * Gets a random lowest card of the given suit from the playable cards.
     */
    public Card getRandomLowestCardOfSuit(Card.Suit suit)
    {
        return CardUtilities.getRandomCard(map.get(suit).get(map.get(suit).firstKey()));
    }

    /**
     * Gets a list of such suits that there is a playable card of each these suits.
     */
    public List<Card.Suit> getListOfSuitsThatIHave()
    {
        return new ArrayList<>(map.keySet());
    }

    /**
     * From {@link #nonLostCards} selects the ones which are of the given suit.
     */
    public List<Card> getNonLostCardsOfGivenSuit(Card.Suit suit)
    {
        return nonLostCards.stream().filter(card -> card.getSuit().equals(suit)).collect(Collectors.toList());
    }

    /**
     * Gets a random playable card that has a suit distinct from spades.
     */
    public Card getRandomCardDistinctFromSpades()
    {
        return CardUtilities.getRandomCard(CardUtilities.getAllCardsWithoutGivenSuit(playableCards, Card.Suit.SPADES));
    }

    /**
     * Gets a random lowest playable spades card that is distinct from the queen of spades. If there is no such card,
     * it'll return <i>null</i>.
     */
    public Card getRandomLowestSpadeDistinctFromQueen()
    {
        Card queen = CardDeck.getInstance().getQueenOfSpades();

        if (isThereCardOfTheSameSuitLowerThan(queen))
            return getRandomLowestCardOfSuit(Card.Suit.SPADES);

        return getRandomLowestCardOfSuitHigherThan(queen);
    }

    /**
     * Gets a random highest card from the given card list.
     */
    public Card getRandomHighestCardFromList(List<Card> cardList)
    {
        SortedMap<Card.Suit, List<Card>> availableCards = CardUtilities.getCardMapFromList(cardList);

        List<Card> highestCards = new ArrayList<>();

        // We'll choose the highest card for each suit.
        for (Card.Suit suit : availableCards.keySet())
        {
            List<Card> cards = availableCards.get(suit);
            List<Card> highestCardsOfGivenSuit = getHighestCardsFromCardList(cards);
            highestCards.addAll(highestCardsOfGivenSuit);
        }

        return CardUtilities.getRandomCard(highestCards);
    }

    /**
     * Gets the highest cards from the cards contained in the given card list. We assume that all the cards in the card
     * list have the same suit.
     */
    private List<Card> getHighestCardsFromCardList(List<Card> cardList)
    {
        SortedMap<Integer, List<Card>> mapForSuit = map.get(cardList.get(0).getSuit());

        int maximum = 0;
        List<Card> result = new ArrayList<>();

        // All the cards in the card list should be in our map for the suit of these cards. We'll go through the whole
        // map.
        for (Integer rank : mapForSuit.keySet())
        {
            List<Card> cardsOfRank = mapForSuit.get(rank);

            // We'll find the intersection between cards of the current rank and the given card list.
            List<Card> copyOfCardList = new ArrayList<>(cardList);
            copyOfCardList.retainAll(cardsOfRank);

            // If the intersection is empty, we'll skip the rank.
            if (copyOfCardList.isEmpty())
                continue;

            // Otherwise we have found such a rank that there are some cards from the card list that have that rank. If
            // this rank is greater than the current maximum, we'll update the maximum.
            if (rank > maximum)
            {
                maximum = rank;
                result = copyOfCardList;
            }
        }

        return result;
    }

    /**
     * From the {@link #nonLostCards} first selects the card suits of which there are the most cards left in the game
     * and then returns a random lowest card of these suits.
     */
    public Card getRandomLowestCardOfSuitWithMostCardsLeftInGameFromNotLostCards()
    {
        List<Card.Suit> suitsFromCards = CardUtilities.getListOfCardSuits(nonLostCards);
        List<Card.Suit> bestSuits = getListOfSuitsWithMaximumNumberOfCardsLeftInGame(suitsFromCards);

        List<Card> cards = new ArrayList<>();

        for (Card.Suit suit : bestSuits)
        {
            List<Card> lowestCardsOfGivenSuit = map.get(suit).get(map.get(suit).firstKey());
            cards.addAll(lowestCardsOfGivenSuit);
        }

        return CardUtilities.getRandomCard(cards);
    }

    /**
     * From the given list of card suits selects the card suits of which there are the most cards left in the game.
     */
    private List<Card.Suit> getListOfSuitsWithMaximumNumberOfCardsLeftInGame(List<Card.Suit> suits)
    {
        List<Card.Suit> result = new ArrayList<>();

        int maximum = 0;

        for (Card.Suit suit : suits)
        {
            int numberOfCardsLeft = getNumberOfCardsOfGivenSuitLeftInGame(suit);

            if (numberOfCardsLeft > maximum)
            {
                maximum = numberOfCardsLeft;
                result.clear();
                result.add(suit);
            }
            else if (numberOfCardsLeft == maximum)
                result.add(suit);
        }

        return result;
    }

    /**
     * Gets a random highest card of the suit as the given one, lower than the given one. If there is no such card,
     * it'll return <i>null</i>.
     */
    public Card getRandomHighestCardOfGivenSuitLessThan(Card card)
    {
        SortedMap<Integer, List<Card>> mapOfGivenSuit = map.get(card.getSuit());

        // if we don't have this suit, we'll return null
        if (mapOfGivenSuit == null)
            return null;

        // this will find mapping for cards lower than the given one
        SortedMap<Integer, List<Card>> subMap = mapOfGivenSuit.subMap(0, card.getRank());

        // if there is no higher card, we'll return null
        if (subMap.isEmpty())
            return null;

        return CardUtilities.getRandomCard(mapOfGivenSuit.get(subMap.lastKey()));
    }

    /**
     * Gets a random lowest card of the suit as the given one, higher than the given one. If there is no such card,
     * it'll return <i>null</i>.
     */
    public Card getRandomLowestCardOfSuitHigherThan(Card card)
    {
        SortedMap<Integer, List<Card>> mapOfGivenSuit = map.get(card.getSuit());

        // if we don't have this suit, we'll return null
        if (mapOfGivenSuit == null)
            return null;

        // this will find mapping for cards higher than the given one
        SortedMap<Integer, List<Card>> subMap = mapOfGivenSuit.subMap(card.getRank() + 1, Card.ACE_RANK + 1);

        // if there is no higher card, we'll return null
        if (subMap.isEmpty())
            return null;

        return CardUtilities.getRandomCard(mapOfGivenSuit.get(subMap.firstKey()));
    }

    /**
     * Finds the number of cards of the given suit that have not been played yet.
     */
    public int getNumberOfCardsOfGivenSuitLeftInGame(Card.Suit suit)
    {
        int numberOfCardsLeft = 0;

        for (Card card : CardDeck.getInstance().getListOfCardsOfSuit(suit))
            if (!playableCards.contains(card) && !playedCardsWithoutTrickCards.contains(card))
                numberOfCardsLeft++;

        return numberOfCardsLeft;
    }

    /**
     * Finds the number of cards that are currently playable of the suit of the given card, less than the given card.
     */
    public int getNumberOfCardsOfTheSuitAsGivenLessThanGiven(Card card)
    {
        SortedMap<Integer, List<Card>> mapOfGivenSuit = map.get(card.getSuit());

        // if we don't have this suit, we'll return 0
        if (mapOfGivenSuit == null)
            return 0;

        // this will find mapping for cards less than the given one
        SortedMap<Integer, List<Card>> subMap = mapOfGivenSuit.subMap(0, card.getRank());

        int result = 0;

        for (List<Card> list : subMap.values())
            result += list.size();

        return result;
    }

    /**
     * From the given list of card suits selects the ones which I have the least playable cards from and returns the
     * playable cards of these suits.
     */
    public List<Card> getListOfCardsOfSuitWhichIHaveTheLeast(List<Card.Suit> availableSuits)
    {
        List<Card> availableCards = new ArrayList<>();
        List<Card.Suit> suits = getListOfSuitsWhichIHaveTheLeastFrom(availableSuits);

        for (Card.Suit suit : suits)
            availableCards.addAll(CardUtilities.getListOfCardsOfGivenSuit(playableCards, suit));

        return availableCards;
    }

    /**
     * From the given list of card suits selects the ones which I have the least playable cards from.
     */
    private List<Card.Suit> getListOfSuitsWhichIHaveTheLeastFrom(List<Card.Suit> availableSuits)
    {
        List<Card.Suit> bestSuits = new ArrayList<>();

        int minimum = Integer.MAX_VALUE;

        for (Card.Suit suit : availableSuits)
        {
            final int number = CardUtilities.getListOfCardsOfGivenSuit(playableCards, suit).size();

            if (number < minimum)
            {
                bestSuits.clear();
                bestSuits.add(suit);
                minimum = number;
            }
            else if (number == minimum)
                bestSuits.add(suit);
        }

        return bestSuits;
    }
}