package gameElements.utilities;

import gameElements.*;

import java.util.ArrayList;
import java.util.List;
import java.util.SortedMap;

/**
 * Provides static utilities for the game elements, especially the method {@link #getPlayableCards(List, Round)}.
 */
public class GameUtilities
{
    /**
     * Chooses currently playable cards, from a given list of cards (assumed to be current hand cards of a player),
     * according to a given round.
     */
    public static List<Card> getPlayableCards(List<Card> cardList, Round round)
    {
        List<Card> result = new ArrayList<>();

        // if a player has the starting card (by default 2 of clubs), then they can play only this card
        Card startingCard = CardDeck.getInstance().getTwoOfClubs();

        if (cardList.contains(startingCard))
        {
            result.add(startingCard);
            return result;
        }

        // let's assume that the player is a starter
        if (round.getCurrentTrick().isEmpty())
        {
            // they surely can play all not-hearts
            result.addAll(CardUtilities.getAllCardsWithoutGivenSuit(cardList, Card.Suit.HEARTS));

            // they can play hearts if and only if hearts are broken, or they have no other option
            if (CardUtilities.hasHeartsOnly(cardList) || round.areHeartsBroken())
                result.addAll(CardUtilities.getListOfCardsOfGivenSuit(cardList, Card.Suit.HEARTS));

            return result;
        }

        // at this point the player is not a starter. If there is a card of the suit of the first played card, a
        // the player can play only cards of this suit
        SortedMap<Card.Suit, List<Card>> cardMap = CardUtilities.getCardMapFromList(cardList);
        Card.Suit suitOfFirst = round.getCurrentTrick().getPlayedCardsList().get(0).getSuit();

        if (cardMap.containsKey(suitOfFirst))
        {
            result.addAll(cardMap.get(suitOfFirst));
            return result;
        }

        // let's assume that the player doesn't have a card of the suit of the first. If they have only hearts,
        // then they basically can play every card. Also if the current round isn't the first.
        if (!round.isTheCurrentTrickFirst() || CardUtilities.hasHeartsOnly(cardList))
        {
            result.addAll(cardList);
            return result;
        }

        // the current round is first and we have a card without penalty to play, therefore we an play only cards
        // without penalty
        result.addAll(CardUtilities.getCardsWithZeroPenalty(cardList));
        return result;
    }

    /**
     * Finds player who did not play the current trick.
     *
     * @param players             A list of all players.
     * @param indexOfStarter      An index of player who started the trick (or who is going to start).
     * @param numberOfPlayedCards Number of cards that've been already played.
     *
     * @return A list of players, in the order in which they're going to play.
     */
    static List<Player> getPlayersWhoDidNotPlayTrick(List<Player> players, int indexOfStarter, int numberOfPlayedCards)
    {
        List<Player> result = new ArrayList<>();

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS - numberOfPlayedCards; i++)
            result.add(players.get((indexOfStarter + numberOfPlayedCards + i) % Game.NUMBER_OF_PLAYERS));

        return result;
    }
}