package gameElements.utilities;

import gameElements.Card;
import gameElements.Player;
import gameElements.Round;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A helper class, which in a round keeps track of suits that particular players can or can't have. The class doesn't
 * know the player's cards themselves. It can be imagine as an observer of the round, who can't see the players cards,
 * but has a perfect memory and based on the game rules can objectively tell which player can / can't have a given
 * suit.
 */
public class SuitWatcher
{
    /**
     * The round that uses the watcher.
     */
    private final Round round;

    /**
     * The mapping of players to the map of suit to the boolean, which indicates, whether player can or can't have the
     * particular suit.
     */
    private final Map<Player, Map<Card.Suit, Boolean>> mainMap;

    /**
     * Construct a new suit watcher for a given round.
     */
    public SuitWatcher(Round round)
    {
        this.round = round;
        mainMap = new HashMap<>();
        initializeMap();
    }

    /**
     * Initialize the main map. At the beginning, each player can have every suit.
     */
    private void initializeMap()
    {
        Map<Card.Suit, Boolean> playerMap = new HashMap<>();

        for (Card.Suit suit : Card.Suit.getValues())
            playerMap.put(suit, true);

        for (Player player : round.getPlayers())
            mainMap.put(player, new HashMap<>(playerMap));
    }

    /**
     * For the given player sets that they haven't got a given suit.
     */
    private void playerDoesNotHaveTheSuit(Player player, Card.Suit suit)
    {
        mainMap.get(player).put(suit, false);
    }

    /**
     * For all players sets that they don't have a given suit.
     */
    private void nobodyHasTheSuit(Card.Suit suit)
    {
        for (Player player : mainMap.keySet())
            playerDoesNotHaveTheSuit(player, suit);
    }

    /**
     * For a given player sets that they have only a given suit.
     */
    private void playerDoesNotHaveSuitsDifferentFrom(Player playerOnTurn, Card.Suit suit)
    {
        for (Card.Suit cs : Card.Suit.getValues())
            if (!cs.equals(suit))
                playerDoesNotHaveTheSuit(playerOnTurn, cs);
    }

    /**
     * Updates the watcher according to the current round state. It's assumed that the current trick is not empty when
     * the method is called. It's also assumed that the method will be called after each move, exactly after we add the
     * played card to the trick list and to the played cards list.
     */
    public void updateMap()
    {
        List<Card> trickCards = round.getCurrentTrick().getPlayedCardsList();
        List<Card> playedCards = round.getPlayedCardsList();

        Card.Suit suitOfFirstInTrick = trickCards.get(0).getSuit();
        Card.Suit suitOfLastPlayed = trickCards.get(trickCards.size() - 1).getSuit();

        Player currentPlayer = round.getPlayerOnTurn();
        boolean heartsAreBroken = round.areHeartsBroken();

        // if the last played card was the last card from it's suit, then we're sure that no player
        // would have it
        if (CardUtilities.areThereAllPossibleCardsOfGivenSuit(playedCards, suitOfLastPlayed))
        {
            nobodyHasTheSuit(suitOfLastPlayed);
            return;
        }

        // if some player started the trick with hearts and hearts hadn't been broken before,
        // he must not have had another option, so he has had only hearts
        if (trickCards.size() == 1 && suitOfLastPlayed.equals(Card.Suit.HEARTS) && !heartsAreBroken)
        {
            playerDoesNotHaveSuitsDifferentFrom(currentPlayer, Card.Suit.HEARTS);
            return;
        }

        // if some player played a card which's suit differs from the suit of the first played card, then
        // they surely don't have the suit of the first played card
        if (!suitOfLastPlayed.equals(suitOfFirstInTrick))
            playerDoesNotHaveTheSuit(currentPlayer, suitOfFirstInTrick);
    }

    /**
     * Gets if a given player could have a given suit.
     */
    boolean couldPlayerHaveSuit(Player player, Card.Suit suit)
    {
        return mainMap.get(player).get(suit);
    }
}