package repositories;

import gameElements.Game;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

/**
 * Provides methods for saving and loading the game statistics and settings. The class is designed as a singleton.
 */
public class GameRepository
{
    /**
     * The instance of the repository.
     */
    private static GameRepository instance;

    /**
     * The file where the game stats are stored in.
     */
    private File statsFile;

    /**
     * The file where the game settings are stored in.
     */
    private File settingsFile;

    /**
     * Constructs the game repository.
     */
    private GameRepository()
    {
        try
        {
            statsFile = IOUtilities.getAndCreateFileInLocalAppFolder("stats.dat");
            settingsFile = IOUtilities.getAndCreateFileInLocalAppFolder("settings.dat");
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    /**
     * Gets the current stats from the {@link #statsFile}.
     */
    private GameStatistics getCurrentStats()
    {
        try (Scanner sc = new Scanner(statsFile))
        {
            if (sc.hasNextLine())
                return GameStatistics.getInstanceFromString(sc.nextLine());

            return GameStatistics.getEmptyStats();
        }
        catch (FileNotFoundException e)
        {
            System.err.println("Couldn't load the file " + statsFile);
            return null;
        }
        catch (RuntimeException e)
        {
            System.err.println(e.getMessage());
            return null;
        }
    }

    /**
     * Records the given stats to the {@link #statsFile}.
     */
    private void recordStats(GameStatistics stats)
    {
        try (PrintWriter pw = new PrintWriter(statsFile))
        {
            pw.println(stats);
        }
        catch (IOException e)
        {
            System.err.println("Couldn't write into the file " + statsFile);
        }
    }

    /**
     * Records the given game from the perspective of the player with the given index.
     */
    public void recordGame(Game game, int indexOfPlayer)
    {
        GameStatistics currentStats = getCurrentStats();

        if (currentStats == null)
            throw new RuntimeException("Could not load the stats.");

        currentStats.setGamesPlayed(currentStats.getGamesPlayed() + 1);

        if (game.getIndexOfWinner() == indexOfPlayer)
            currentStats.setGamesWon(currentStats.getGamesWon() + 1);

        final int currentGameRounds = game.getCurrentRoundNumber();
        final int currentGamePoints = game.getPlayers().get(indexOfPlayer).getCurrentGameScore();
        final int currentGameMoonshots = game.getPlayers().get(indexOfPlayer).getNumberOfMoonshots();

        currentStats.setTotalRounds(currentStats.getTotalRounds() + currentGameRounds);
        currentStats.setTotalPoints(currentStats.getTotalPoints() + currentGamePoints);
        currentStats.setMoonshots(currentStats.getMoonshots() + currentGameMoonshots);

        recordStats(currentStats);
    }

    /**
     * Resets the stats to its default (zero) values.
     */
    public void resetStats()
    {
        recordStats(GameStatistics.getEmptyStats());
    }

    /**
     * Gets the stats as a list of strings which should be written into the statistics screen table.
     */
    public List<String> getCurrentStatsAsListOfStrings()
    {
        List<String> result = new ArrayList<>();

        GameStatistics stats = getCurrentStats();

        if (stats == null)
            throw new RuntimeException("Could not load the stats.");

        result.add(Integer.toString(stats.getGamesPlayed()));
        result.add(Integer.toString(stats.getGamesWon()));

        int winningPercentage = 0;

        if (stats.getGamesPlayed() != 0)
            winningPercentage = (int) (100 * (double) stats.getGamesWon() / stats.getGamesPlayed());

        result.add(Integer.toString(winningPercentage) + " %");
        result.add(Integer.toString(stats.getTotalRounds()));

        int avgPointsPerRound = 0;

        if (stats.getTotalRounds() != 0)
            avgPointsPerRound = stats.getTotalPoints() / stats.getTotalRounds();

        result.add(Integer.toString(avgPointsPerRound));
        result.add(Integer.toString(stats.getMoonshots()));

        return result;
    }

    /**
     * Gets the music state from the settings file. If reading of the file doesn't succeed, it'll return <i>true.</i>
     */
    public boolean getMusicState()
    {
        try (Scanner sc = new Scanner(settingsFile))
        {
            return !sc.hasNextLine() || Boolean.parseBoolean(sc.nextLine());
        }
        catch (FileNotFoundException e)
        {
            System.err.println("Couldn't load the file " + settingsFile);
            return true;
        }
    }

    /**
     * Saves the music state given by the parameter to the settings file.
     */
    public void saveMusicState(boolean isMusicOn)
    {
        try (PrintWriter pw = new PrintWriter(settingsFile))
        {
            pw.println(isMusicOn);
        }
        catch (IOException e)
        {
            System.err.println("Couldn't write into the file " + settingsFile);
        }
    }

    /**
     * Gets the instance of the repository.
     */
    public static GameRepository getInstance()
    {
        if (instance == null)
            instance = new GameRepository();

        return instance;
    }
}