package repositories;

import gameElements.Card;
import sk.upjs.jpaz2.ImageShape;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Consumer;
import java.util.function.Function;

/**
 * Provides methods for images saving and loading. The class is designed as a singleton.
 */
public class ImageRepository
{
    /**
     * The instance of the repository.
     */
    private static ImageRepository instance;

    /**
     * The mapping of the loaded resource shapes to its image shapes.
     */
    private final Map<String, ImageShape> loadedShapes;

    /**
     * Constructs the image repository.
     */
    private ImageRepository()
    {
        loadedShapes = new HashMap<>();
    }

    /**
     * Gets the image shape from the resource file name, stored in the package {@link resImages} or {@link
     * resImagesForBuild}.
     */
    public ImageShape getImageShape(String resourceName)
    {
        if (loadedShapes.containsKey(resourceName))
            return loadedShapes.get(resourceName);

        throw new RuntimeException("Couldn't get the resource: " + resourceName);
    }

    /**
     * Gets the image shape which represents the given card. If the card is null, it will return the image shape of the
     * card from back.
     */
    public ImageShape getImageShape(Card card)
    {
        try
        {
            return card == null ? getImageShape("plainCard.png") : getImageShape(card.toString() + ".png");
        }
        catch (RuntimeException e)
        {
            System.err.println("Couldn't get the card: " + card);
            throw e;
        }
    }

    /**
     * Loads all the images from the packages {@link resImages}, {@link resImagesForBuild} and {@link resImagesCards} to
     * the memory.
     */
    public void loadAllImages()
    {
        loadImages();
        loadImagesForBuild();
    }

    /**
     * Loads all the simple image shapes (those view just one view) from the packages {@link resImages}, {@link
     * resImagesCards}) to the memory.
     */
    private void loadImages()
    {
        Function<String, Consumer<String>> action = packageName -> name ->
        {
            try
            {
                ImageShape img = new ImageShape(packageName, name);
                loadedShapes.put(name, img);
            }
            catch (Exception e)
            {
                System.err.println("Couldn't load the image " + name);
            }
        };

        try
        {
            IOUtilities.executeForAllFilesInResourceFolder("resImagesCards", action.apply("resImagesCards"));
            IOUtilities.executeForAllFilesInResourceFolder("resImages", action.apply("resImages"));
        }
        catch (IOException e)
        {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Loads all the image shapes with 2 views, from the package {@link resImagesForBuild}.
     */
    private void loadImagesForBuild()
    {
        try
        {
            IOUtilities.executeForAllFilesInResourceFolder("resImagesForBuild", name ->
            {
                try
                {
                    ImageShape.Builder imageBuilder = new ImageShape.Builder("resImagesForBuild", name);
                    imageBuilder.setViewCount(2);
                    loadedShapes.put(name, imageBuilder.createShape());
                }
                catch (Exception e)
                {
                    System.err.println("Couldn't load the image " + name);
                }
            });
        }
        catch (IOException e)
        {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Gets the instance of the repository.
     */
    public static ImageRepository getInstance()
    {
        if (instance == null)
            instance = new ImageRepository();

        return instance;
    }
}