package userInterface;

import gameElements.Game;
import gameElements.Card;
import repositories.GameRepository;
import userInterface.utilities.TimerUtilities;
import userInterface.utilities.TimerUtilities.Timer;

import java.util.List;

/**
 * A class, which provides communication between {@link GameScreen} and game elements. For example, if someone plays a
 * card, the screen will find out the card, send it to the initializer, which will send it to the current round and also
 * process the answer from it. This class helps us to distinct the viewing logic from the playing logic.
 */
class GameInitializer
{
    /**
     * The game screen which uses the initializer.
     */
    private final GameScreen gameScreen;

    /**
     * Indicates if the current mode is the playing mode. If it's false, the mode is the changing mode.
     */
    private boolean playingMode;

    /**
     * Constructs a new initializer that is going to be used by a given game screen.
     */
    GameInitializer(GameScreen gameScreen)
    {
        this.gameScreen = gameScreen;
    }

    /**
     * Prepares and starts a next round.
     */
    void startNextRound()
    {
        getGame().prepareNewRound();
        gameScreen.prepareNextRound();
    }

    /**
     * Prepares changes for the current round.
     */
    void prepareChanges()
    {
        boolean changeRound = !getGame().getCurrentRound().isThisNoChangeRound();
        playingMode = !changeRound;
        gameScreen.prepareForChanges(changeRound);
    }

    /**
     * Executes the given changes of the current round.
     *
     * @param forChange The cards for change of the human player.
     */
    void executeChanges(List<Card> forChange)
    {
        getGame().getCurrentRound().passCards(GameScreen.INDEX_OF_HUMAN, forChange);

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            if (i == GameScreen.INDEX_OF_HUMAN)
                continue;

            List<Card> list = getGame().getPlayers().get(i).getPassSuggestion();
            getGame().getCurrentRound().passCards(i, list);
        }

        List<Card> received = getGame().getCurrentRound().getReceivedCardsOfPlayer(GameScreen.INDEX_OF_HUMAN);
        gameScreen.showReceivedCards(received);
    }

    /**
     * Starts the next trick.
     */
    void startNextTrick()
    {
        playingMode = true;

        if (isHumanOnTurn())
        {
            gameScreen.humanOnTurn();
            return;
        }

        executeMoves();
    }

    /**
     * Execute moves, until the human player is on turn.
     */
    private void executeMoves()
    {
        // there will be at most Game.NUMBER_OF_PLAYERS ticks. Maybe it will stop earlier
        Timer timer = TimerUtilities.getPlainPeriodicalTimer(GameScreen.TIME_BETWEEN_MOVES, Game.NUMBER_OF_PLAYERS);

        timer.setConsumer(i ->
        {
            // if the current trick is over, we turn the timer off
            if (isCurrentTrickOver())
            {
                getGame().getCurrentRound().prepareNewTrick();
                gameScreen.trickIsOver();
                timer.setEnabled(false);
                return;
            }

            // if the current trick isn't over and the human is on turn, we turn
            // the timer off as well
            if (isHumanOnTurn())
            {
                gameScreen.humanOnTurn();
                timer.setEnabled(false);
                return;
            }

            // otherwise computer players are on turn, we execute their moves
            executeComputerMove();
        });

        timer.setEnabled(true);
    }

    /**
     * Executes a computer move. It automatically finds out who is on turn. It's supposed that it's not the human
     * player.
     */
    private void executeComputerMove()
    {
        int index = getGame().getCurrentRound().getIndexOfPlayerOnTurn();
        Card card = getGame().getPlayers().get(index).getMoveSuggestion();
        getGame().getCurrentRound().playNextMove(card);
        gameScreen.displayCardPlayedByComputer(card, index);
    }

    /**
     * Executes a human move.
     *
     * @param card The card that the human player is going to play.
     */
    void executeHumanMove(Card card)
    {
        getGame().getCurrentRound().playNextMove(card);
        executeMoves();
    }

    /**
     * Handles the situation after clearing the trick. If the current round isn't over, it'll start a new trick.
     * Otherwise it will display the {@link RoundResultScreen} and if the game is over, also save the game.
     */
    void afterClearingTrick()
    {
        if (getGame().getCurrentRound().hasRoundEnded())
        {
            gameScreen.displayRoundResultScreen();

            if (getGame().hasGameEnded())
                GameRepository.getInstance().recordGame(getGame(), GameScreen.INDEX_OF_HUMAN);

            return;
        }

        startNextTrick();
    }

    /**
     * Finds out whether the current trick is over.
     */
    private boolean isCurrentTrickOver()
    {
        return getGame().getCurrentRound().getCurrentTrick().isFull();
    }

    /**
     * Finds out whether the human is on turn.
     */
    private boolean isHumanOnTurn()
    {
        return getGame().getCurrentRound().isPlayerOnTurn(GameScreen.INDEX_OF_HUMAN);
    }

    /**
     * Gets the game which belongs to the game screen which uses this initializer.
     */
    private Game getGame()
    {
        return gameScreen.getGame();
    }

    /**
     * Gets if the current mode is playing (true) or changing (false).
     */
    boolean isCurrentModePlaying()
    {
        return playingMode;
    }
}