package userInterface;

import gameElements.Card;
import gameElements.Game;
import repositories.AudioRepository;
import repositories.ImageRepository;
import sk.upjs.jpaz2.AudioClip;
import sk.upjs.jpaz2.ImageShape;
import userInterface.panes.*;
import userInterface.panes.Button;
import userInterface.panes.Label;
import userInterface.utilities.PaneUtilities;
import userInterface.utilities.TimerUtilities;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.util.*;
import java.util.List;
import java.util.function.Consumer;
import java.util.stream.Collectors;

/**
 * The main screen of the program, which represents a game. The whole game is going to be played in this screen. It is
 * meant that in this class there's no communication between the game elements (only get requests). Every such
 * communication should be done by {@link GameInitializer}.
 */
class GameScreen extends Screen
{
    /**
     * The font of the screen.
     */
    private static final Font SCREEN_FONT = new Font("Calibri", Font.CENTER_BASELINE, 25);

    /**
     * The font color of the screen.
     */
    private static final Color FONT_COLOR = Color.white;

    /**
     * The index of a human player in the player's list of the game.
     */
    static final int INDEX_OF_HUMAN = 0;

    /**
     * The time in milliseconds between moves of computer players.
     */
    static final int TIME_BETWEEN_MOVES = 400;

    /**
     * The game which is displayed by the screen.
     */
    private final Game game;

    /**
     * The game initializer for the game.
     */
    private final GameInitializer gameInitializer;

    /**
     * The list of buttons with player's names.
     */
    private final List<Button> nameButtons;

    /**
     * The list of buttons with player's scores.
     */
    private final List<Button> scoreButtons;

    /**
     * The button with the message "Submit", used for a confirmation of cards for change.
     */
    private TwoStatedButton submitButton;

    /**
     * The back icon, which brings us back to the intro screen.
     */
    private final BackIcon backIcon;

    /**
     * The list of card panes played in the current trick.
     */
    private final List<CardPane> cardPanesOfCurrentTrick;

    /**
     * The map of the player's indices to their card panes.
     */
    private Map<Integer, List<CardPane>> cards;

    /**
     * The map of the player's indices to their card panes for change.
     */
    private Map<Integer, List<CardPane>> cardsForChange;

    /**
     * Indicates whether the screen reacts on onMousePressed event (except for clicking on the back icon).
     */
    private boolean clickable;

    /**
     * Indicates whether we are waiting for press the submit button.
     */
    private boolean waitingForSubmit;

    /**
     * The current X coordinate. For the reason why we need to store it in the variable, see {@link
     * #turnOnClickable()}.
     */
    private int currentX;

    /**
     * The current X coordinate. For the reason why we need to store it in the variable, see {@link
     * #turnOnClickable()}.
     */
    private int currentY;

    /**
     * Constructs a game screen which displays a given game.
     */
    GameScreen(Game game)
    {
        this.game = game;
        gameInitializer = new GameInitializer(this);
        cardPanesOfCurrentTrick = new ArrayList<>();
        nameButtons = new ArrayList<>();
        scoreButtons = new ArrayList<>();

        PaneUtilities.drawBackground(this, "gameScreen.png");
        drawMusicIcon();
        backIcon = new BackIcon();
        backIcon.drawToScreen(this);

        gameInitializer.startNextRound();
    }

    /**
     * Draws or redraws buttons with player's names.
     */
    private void drawNames()
    {
        nameButtons.forEach(this::remove);
        nameButtons.clear();

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            String name = game.getPlayers().get(i).toString();
            Label label = new Label(name, SCREEN_FONT, FONT_COLOR);
            ImageShape shape = ImageRepository.getInstance().getImageShape("nameButton.png");
            Button button = new Button(shape, label);
            nameButtons.add(button);
            add(button);
            button.setPosition(getPositionOfPlayerNameButton(i));
        }
    }

    /**
     * Draws or redraws buttons with player's current scores.
     */
    private void drawScores()
    {
        scoreButtons.forEach(this::remove);
        scoreButtons.clear();

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            String score = Integer.toString(game.getPlayers().get(i).getCurrentGameScore());
            Label label = new Label(score, SCREEN_FONT, FONT_COLOR);
            ImageShape shape = ImageRepository.getInstance().getImageShape("scoreButton.png");
            Button button = new Button(shape, label);
            scoreButtons.add(button);
            add(button);
            button.setPosition(getPositionOfPlayerScoreButton(i));
        }
    }

    /**
     * Draws all cards of player with the given index. It is supposed that cards have been dealt.
     */
    private void drawCardsOfPlayer(int indexOfPlayer)
    {
        List<CardPane> playersCards = cards.get(indexOfPlayer);

        for (int i = 0; i < playersCards.size(); i++)
        {
            CardPane cp = playersCards.get(i);
            remove(cp);
            add(cp);
            bringToBackOf(cp, nameButtons.get(indexOfPlayer));
            cp.setPosition(getPositionOfCard(i, indexOfPlayer, playersCards.size()));
        }
    }

    /**
     * Gets the position of the card with given parameters. Cards are meant to be centered, no matter how many they are.
     * That's why we need the numberOfCards parameter.
     *
     * @param indexOfCard   The index of the card in player's hand.
     * @param indexOfPlayer The index of the player.
     * @param numberOfCards The number of the cards which the player currently has.
     */
    private Point2D getPositionOfCard(int indexOfCard, int indexOfPlayer, int numberOfCards)
    {
        final int distance = (indexOfPlayer % 2 == 0) ? 40 : 25;
        final int dxTranslation = (indexOfPlayer % 2 == 0) ? distance : 0;
        final int dyTranslation = (indexOfPlayer % 2 == 0) ? 0 : distance;

        CardPane pane = cards.get(indexOfPlayer).get(0);
        Point finalPosition = getPositionOfFirstCardInHand(pane, indexOfPlayer, distance, numberOfCards);
        finalPosition.translate(dxTranslation * indexOfCard, dyTranslation * indexOfCard);

        return finalPosition;
    }

    /**
     * Draws the submit button.
     */
    private void drawSubmitButton()
    {
        String text = game.getCurrentRound().getTheWayOfPassingAsString();
        ImageShape shape = ImageRepository.getInstance().getImageShape("submitButton.png");
        Label label = new Label(text, SCREEN_FONT, Color.white);
        submitButton = new TwoStatedButton(shape, label);
        submitButton.setReactive(false);
        submitButton.turnOn();
        add(submitButton);
        submitButton.setPosition(370, 270);
    }

    /**
     * Displays a round result screen.
     */
    void displayRoundResultScreen()
    {
        RoundResultScreen screen = new RoundResultScreen(this);
        Hearts.getInstance().changeScreen(screen, false);
        // we reset the current position because if we don't do that, it might happen that if we
        // return to this screen, some card might be raised, even if the cursor isn't over it
        currentX = 0;
        currentY = 0;
    }

    /**
     * Prepares a new round, i.e. draw player's names, score, initialize cards and deal them.
     */
    void prepareNextRound()
    {
        clickable = false;
        drawNames();
        drawScores();
        initializeCards();
        initializeCardsForChange();
        dealCards();
    }

    /**
     * Initialize playing cards map. It fills it with empty cards.
     */
    private void initializeCards()
    {
        cards = new HashMap<>();

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            List<CardPane> cardsOfPlayer = new ArrayList<>();

            for (int j = 0; j < Game.NUMBER_OF_CARDS_IN_HAND; j++)
                cardsOfPlayer.add(new CardPane());

            cards.put(i, cardsOfPlayer);
        }
    }

    /**
     * Initialize cards for change map. The map will be empty.
     */
    private void initializeCardsForChange()
    {
        cardsForChange = new HashMap<>();

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
            cardsForChange.put(i, new ArrayList<>());
    }

    /**
     * Gets the list of human card panes. We can specify whether we want to have cards ordered or not.
     */
    private List<CardPane> getHumanCardPanes(boolean ordered)
    {
        List<CardPane> result = game.getPlayers().get(INDEX_OF_HUMAN)
                .getCurrentHand().getHandCardsList().stream()
                .map(CardPane::new).collect(Collectors.toList());

        if (!ordered)
            Collections.shuffle(result);

        return result;
    }

    /**
     * Animates dealing cards.
     */
    private void dealCards()
    {
        // first we bring all card panes into the middle of the screen (approximately). The way we do it
        // makes us sure that the panes are in the right order

        for (Integer i : cards.keySet())
        {
            for (CardPane cp : cards.get(i))
            {
                add(cp);
                cp.setPosition((getWidth() - cp.getWidth()) / 2 + 10, (getHeight() - cp.getHeight()) / 2 - 20);
                // we want cards to be under name buttons
                bringToBackOf(cp, nameButtons.get(i));
            }
        }

        // then we need list of the panes ordered in the way that they're going to fly.

        List<CardPane> panes = new ArrayList<>();

        for (int i = 0; i < Game.NUMBER_OF_CARDS_IN_HAND; i++)
            for (int j = 0; j < Game.NUMBER_OF_PLAYERS; j++)
                panes.add(cards.get(j).get(i));

        // we'll first display human card panes in a random order, then we'll animate ordering of them

        List<CardPane> shuffled = getHumanCardPanes(false);

        Consumer<Integer> periodicalAction = i ->
        {
            CardPane c = panes.get(i);
            Point2D position = getPositionOfCard(i / 4, i % 4, Game.NUMBER_OF_CARDS_IN_HAND);
            PaneAnimator pa = new PaneAnimator();
            pa.moveTo(c, position, 300);

            // we display human's cards
            if (i % 4 == INDEX_OF_HUMAN)
                c.setCard(shuffled.get(i / 4).getCard());

            playSound("move.wav");
        };

        // after dealing we reorder human cards for an effect and let initializer prepare card changes

        Runnable actionAfter = () -> TimerUtilities.waitFor(700, () ->
        {
            reorderCards();
            gameInitializer.prepareChanges();
        });

        // we wait a little bit until the dealing starts
        TimerUtilities.waitFor(600, () -> TimerUtilities.executePeriodically(80, panes.size(), periodicalAction, actionAfter));
    }

    /**
     * Prepares the screen for changes.
     *
     * @param changeRound Indicates whether the current round is a change round.
     */
    void prepareForChanges(boolean changeRound)
    {
        TimerUtilities.waitFor(1000, () ->
        {
            if (!changeRound)
            {
                clickable = false;
                gameInitializer.startNextTrick();
                return;
            }

            drawSubmitButton();
            turnOnClickable();
            displayComputerChoices();
        });
    }

    /**
     * Animates reordering human card panes.
     */
    private void reorderCards()
    {
        // first we obtain ordered human card panes, different from the ones in cards map
        List<CardPane> rightOrder = getHumanCardPanes(true);

        for (int i = 0; i < Game.NUMBER_OF_CARDS_IN_HAND; i++)
        {
            CardPane c = rightOrder.get(i);
            // adding panes like this makes us sure they'll be added in the right order
            add(c);

            // we are given the CardPane c. We'll find out the index of this card in the unordered map
            int index = -1;

            for (int j = 0; j < Game.NUMBER_OF_CARDS_IN_HAND; j++)
            {
                CardPane cp = cards.get(INDEX_OF_HUMAN).get(j);

                if (cp.getCard().equals(c.getCard()))
                {
                    index = j;
                    break;
                }
            }

            // we bring the new card pane to the position of it's twin. The difference between them is in the
            // layer in which they lie
            c.setPosition(cards.get(INDEX_OF_HUMAN).get(index).getPosition());
        }

        // we remove original unordered card panes. In this moment the ordered ones are at their positions in
        // the right layers
        cards.get(INDEX_OF_HUMAN).forEach(this::remove);

        // we simply animate each card from it's current position to the position where it should be
        for (int i = 0; i < Game.NUMBER_OF_CARDS_IN_HAND; i++)
        {
            CardPane cp = rightOrder.get(i);
            PaneAnimator pa = new PaneAnimator();
            pa.moveTo(cp, getPositionOfCard(i, INDEX_OF_HUMAN, Game.NUMBER_OF_CARDS_IN_HAND), 400);
        }

        playSound("cardChanges.wav");
        cards.put(INDEX_OF_HUMAN, rightOrder);
    }

    /**
     * In a change round, displays random computer choices.
     */
    private void displayComputerChoices()
    {
        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            if (i == INDEX_OF_HUMAN)
                continue;

            List<CardPane> panes = new ArrayList<>(cards.get(i));

            // we find random card panes for change. The way we do it ensures us that they'll be different
            while (panes.size() != Game.NUMBER_OF_CARDS_TO_CHANGE)
                panes.remove((int) (Math.random() * panes.size()));

            cardsForChange.put(i, panes);

            // and now simply "push them out"
            for (CardPane forChange : panes)
            {
                if (i == 1)
                    forChange.setX(forChange.getX() + 10);

                if (i == 2)
                    forChange.setY(forChange.getY() + 10);

                if (i == 3)
                    forChange.setX(forChange.getX() - 10);
            }
        }
    }

    /**
     * Process the given card pane for change. It handles cases: <br> 1. The card's been already selected for change.
     * (it deselects that). <br> 2. The card is the last for change. (it shades all the cards expect the selected ones)
     */
    private void processForChanging(CardPane cp)
    {
        // if we process the pane which was already selected
        if (getHumansCardsForChange().contains(cp))
        {
            playSound("click.wav");
            getHumansCardsForChange().remove(cp);
            // selected cards are always raised
            cp.lower();
            updateCardsAfterDeselectingOneForChange();
            // surely we can't submit changes yet
            submitButton.setReactive(false);
            submitButton.turnOn();
            waitingForSubmit = false;
            return;
        }

        // in this moment we now the card wasn't selected yet, we select and raise it
        getHumansCardsForChange().add(cp);
        cp.raise(12);
        cp.setCanBeRaised(false);
        playSound("click.wav");

        // if we selected the last for change
        if (getHumansCardsForChange().size() == Game.NUMBER_OF_CARDS_TO_CHANGE)
        {
            updateCardsAfterSelectingLastForChange();
            // we can now click on the submit button to confirm out changes
            submitButton.setReactive(true);
            submitButton.turnOff();
            waitingForSubmit = true;
        }
    }

    /**
     * Lights all the card panes expect the ones for change.
     */
    private void updateCardsAfterDeselectingOneForChange()
    {
        for (CardPane cp : cards.get(INDEX_OF_HUMAN))
        {
            if (getHumansCardsForChange().contains(cp))
                continue;

            cp.setCanBeRaised(true);

            if (cp.isShaded())
                cp.light();
        }
    }

    /**
     * Shades all the card panes except the ones for change.
     */
    private void updateCardsAfterSelectingLastForChange()
    {
        for (CardPane pane : cards.get(INDEX_OF_HUMAN))
        {
            pane.setCanBeRaised(false);

            if (getHumansCardsForChange().contains(pane))
                continue;

            pane.shade(0.2f);
        }
    }

    /**
     * Animates displaying the given received cards.
     */
    void showReceivedCards(List<Card> received)
    {
        final int[] pairs = game.getCurrentRound().getPassingPairs();

        Map<Integer, List<CardPane>> newCards = new HashMap<>(cards);

        // first we build the newCards according to the changes. We need it to order card layers
        for (int i = 0; i < pairs.length; i++)
        {
            List<CardPane> sending = cardsForChange.get(pairs[i]);
            List<CardPane> receiving = cardsForChange.get(i);

            for (int j = 0; j < sending.size(); j++)
            {
                CardPane sent = sending.get(j);
                int index = cards.get(i).indexOf(receiving.get(j));
                newCards.get(i).set(index, sent);
            }
        }

        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            for (CardPane cp : newCards.get(i))
            {
                Point2D position = cp.getPosition();
                // removing and adding each card pane makes us sure that new card panes will lie
                // in the right layer
                remove(cp);
                add(cp);
                // we want cards under the name buttons
                bringToBackOf(cp, nameButtons.get(i));
                cp.setPosition(position);
            }
        }

        // now we have cards in the right layers.  We animate each change. Cards which human sent we
        // hide and cards which human gets we show.
        for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
        {
            List<CardPane> sending = cardsForChange.get(pairs[i]);
            List<CardPane> receiving = cardsForChange.get(i);

            for (int j = 0; j < sending.size(); j++)
            {
                CardPane sent = sending.get(j);

                if (pairs[i] == INDEX_OF_HUMAN)
                    sent.setCard(null);

                if (i == INDEX_OF_HUMAN)
                    sent.setCard(received.get(j));

                new PaneAnimator().moveTo(sending.get(j), receiving.get(j).getPosition(), 600);
            }
        }

        playSound("cardChanges.wav");

        TimerUtilities.waitFor(1600, () ->
        {
            // the cards are now "pushed out". We'll bring them to the right position
            for (int i = 0; i < Game.NUMBER_OF_PLAYERS; i++)
            {
                for (int j = 0; j < Game.NUMBER_OF_CARDS_IN_HAND; j++)
                {
                    CardPane cp = newCards.get(i).get(j);
                    cp.setPosition(getPositionOfCard(j, i, Game.NUMBER_OF_CARDS_IN_HAND));
                }
            }

            // after this we reorder cards and start a new trick.
            TimerUtilities.waitFor(600, () ->
            {
                reorderCards();
                TimerUtilities.waitFor(1000, gameInitializer::startNextTrick);
            });
        });
    }

    /**
     * Process the given card for playing.
     */
    private void processForPlaying(CardPane cp)
    {
        hidePlayableCards();
        displayCardPane(cp, INDEX_OF_HUMAN);
        gameInitializer.executeHumanMove(cp.getCard());
    }

    /**
     * Displays the given card to be played by the given player.
     */
    private void displayCardPane(CardPane pane, int playerIndex)
    {
        playSound("move.wav");
        pane.setCanBeRaised(false);
        bringToFront(pane);
        cardPanesOfCurrentTrick.add(pane);
        cards.get(playerIndex).remove(pane);
        new PaneAnimator().moveTo(pane, getPositionOfPlayedCard(playerIndex), TIME_BETWEEN_MOVES);
        drawCardsOfPlayer(playerIndex);
    }

    /**
     * Displays the given card played by computer.
     */
    void displayCardPlayedByComputer(Card card, int playerIndex)
    {
        List<CardPane> playersCards = cards.get(playerIndex);
        CardPane randomPane = playersCards.get((int) (Math.random() * playersCards.size()));
        randomPane.setCard(card);
        displayCardPane(randomPane, playerIndex);
    }

    void trickIsOver()
    {
        int trickWinner = game.getCurrentRound().getIndexOfLastTrickWinner();

        // we wait a little bit in order for player to be able to see the played cards
        TimerUtilities.waitFor(500, () ->
        {
            playSound("tableClear.wav");

            final int moveTime = 400;

            // we animate each pane to the given position
            for (CardPane cp : cardPanesOfCurrentTrick)
                new PaneAnimator().moveTo(cp, getEndPositionOfMovingCard(trickWinner), moveTime);

            // we must wait until the moving is over
            TimerUtilities.waitFor(moveTime, () ->
            {
                cardPanesOfCurrentTrick.forEach(this::remove);
                cardPanesOfCurrentTrick.clear();
                gameInitializer.afterClearingTrick();
            });
        });
    }

    /**
     * Shades all human card's that they can't play.
     */
    private void showPlayableCards()
    {
        List<Card> playableCards = game.getPlayers().get(INDEX_OF_HUMAN).getCurrentlyPlayableCards();

        cards.get(INDEX_OF_HUMAN)
                .stream()
                .filter(cardPane -> !playableCards.contains(cardPane.getCard()))
                .forEach(cardPane -> cardPane.shade(0.2f));
    }

    /**
     * Lights all human card's that they can't play. It's supposed that they've been already shaded.
     */
    private void hidePlayableCards()
    {
        cards.get(INDEX_OF_HUMAN)
                .stream()
                .filter(PaneWithShadow::isShaded)
                .forEach(PaneWithShadow::light);
    }

    /**
     * Shows playable cards and turns on clickable (i.e. allows screen to react on mouse clicked and moved events)
     */
    void humanOnTurn()
    {
        showPlayableCards();
        turnOnClickable();
    }

    /**
     * Turns on clickable, i.e. allows screen to react on mouse clicked and moved events.
     */
    private void turnOnClickable()
    {
        clickable = true;
        // if the cursor is over some card pane, it will raise this card pane. This is one of the reasons for
        // introducing currentX, currentY variables.
        updateStateOfCardPanes();
    }

    /**
     * Returns the unshaded card pane on the current position (given by currentX, currentY variables), or null, if there
     * is no such pane.
     */
    private CardPane getUnshadedCardPaneOnCurrentPosition()
    {
        CardPane result = null;

        for (CardPane cp : cards.get(INDEX_OF_HUMAN))
            if (cp.containsPoint(currentX, currentY))
                result = cp;

        if (result != null && result.isShaded())
            result = null;

        return result;
    }

    /**
     * Changes the screen to this and starts a next round.
     */
    void continueInGame()
    {
        Hearts.getInstance().changeScreen(this, false);
        gameInitializer.startNextRound();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onMousePressed(int x, int y, MouseEvent detail)
    {
        // first we check if we have clicked on the back icon
        if (backIcon.containsPoint(x, y))
        {
            playSound("click.wav");
            turnOffMusic();
            Hearts.getInstance().changeScreen(IntroScreen.getInstance(), true);
            return;
        }

        // if the screen is not clickable now, we return
        if (!clickable)
            return;

        // otherwise we change clickable to false, in order to prevent from weird behaviour
        // caused by fast clicking
        clickable = false;

        // if we've clicked on the submit button and we're waiting for submit
        if (waitingForSubmit && submitButton.containsPoint(x, y))
        {
            playSound("click.wav");
            remove(submitButton);
            waitingForSubmit = false;
            List<Card> changes = getHumansCardsForChange().stream().map(CardPane::getCard).collect(Collectors.toList());
            gameInitializer.executeChanges(changes);
            getHumansCardsForChange().clear();
            return;
        }

        CardPane cp = getUnshadedCardPaneOnCurrentPosition();

        // if we haven't clicked on a card, we could simply turn on clickable and continue
        if (cp == null)
        {
            turnOnClickable();
            return;
        }

        // otherwise we've clicked on a card. If the current mode is playing, we play the card, otherwise we
        // change the card
        if (gameInitializer.isCurrentModePlaying())
        {
            processForPlaying(cp);
        }
        else
        {
            processForChanging(cp);
            turnOnClickable();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onMouseMoved(int x, int y, MouseEvent detail)
    {
        currentX = x;
        currentY = y;

        updateStateOfSubmitButton();

        if (clickable)
            updateStateOfCardPanes();
    }

    /**
     * Updates state of the submit button according to the current position given by the variables currentX, currentY.
     */
    private void updateStateOfSubmitButton()
    {
        if (!PaneUtilities.contains(this, submitButton))
            return;

        if (!submitButton.isReactive())
            return;

        submitButton.turnOff();

        if (submitButton.containsPoint(currentX, currentY))
            submitButton.turnOn();
    }

    /**
     * Updates state of human card panes according to the current position given by the variables currentX, currentY.
     */
    private void updateStateOfCardPanes()
    {
        // we lower all panes that are raised and can be raised (i.e. also lowered)
        cards.get(INDEX_OF_HUMAN).stream()
                .filter(cardPane -> cardPane.isRaised() && cardPane.isCanBeRaised())
                .forEach(CardPane::lower);

        CardPane cp = getUnshadedCardPaneOnCurrentPosition();

        // if the cursor is over a pane that can be raised, we'll raise it
        if (cp != null && cp.isCanBeRaised())
            cp.raise(12);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected boolean onCanClick(int x, int y)
    {
        return super.onCanClick(x, y) || canClickOnCard();
    }

    /**
     * Gets if the cursor should be hand because it's over a clickable card.
     */
    private boolean canClickOnCard()
    {
        if (!clickable)
            return false;

        CardPane cp = getUnshadedCardPaneOnCurrentPosition();

        return cp != null && cp.isRaised();
    }

    /**
     * Gets the position of the first card in hand of the given player.
     *
     * @param card          The given card pane.
     * @param indexOfPlayer The given player's index.
     * @param distance      The distance between cards in the player's hand.
     * @param numberOfCards The total number of cards in the player's hand.
     */
    private Point getPositionOfFirstCardInHand(CardPane card, int indexOfPlayer, int distance, int numberOfCards)
    {
        int a = (getWidth() - card.getWidth() - (numberOfCards - 1) * distance) / 2;
        int b = (getHeight() - card.getHeight() - (numberOfCards - 1) * distance) / 2;

        switch (indexOfPlayer)
        {
            case 0:
                return new Point(a, 556);
            case 1:
                return new Point(54, b);
            case 2:
                return new Point(a, 18);
            case 3:
                return new Point(780, b);
            default:
                throw new IllegalStateException();
        }
    }

    /**
     * Gets the position of a given player's name button.
     */
    private Point getPositionOfPlayerNameButton(int indexOfPlayer)
    {
        switch (indexOfPlayer)
        {
            case 0:
                return new Point(356, 500);
            case 1:
                return new Point(39, 352);
            case 2:
                return new Point(356, 145);
            case 3:
                return new Point(645, 352);
            default:
                throw new IllegalStateException();
        }
    }

    /**
     * Gets the position of a given player's score button.
     */
    private Point getPositionOfPlayerScoreButton(int playerIndex)
    {
        switch (playerIndex)
        {
            case 0:
                return new Point(508, 500);
            case 1:
                return new Point(191, 352);
            case 2:
                return new Point(508, 145);
            case 3:
                return new Point(797, 352);
            default:
                throw new IllegalStateException();
        }
    }

    /**
     * Gets the ending position of a played card by the given player.
     */
    private Point getPositionOfPlayedCard(int playerIndex)
    {
        switch (playerIndex)
        {
            case 0:
                return new Point(409, 356);
            case 1:
                return new Point(340, 263);
            case 2:
                return new Point(409, 189);
            case 3:
                return new Point(482, 263);
            default:
                throw new IllegalStateException();
        }
    }

    /**
     * Gets the ending position of a card moving to the winner after the end of trick.
     */
    private Point getEndPositionOfMovingCard(int winnerIndex)
    {
        CardPane card = cardPanesOfCurrentTrick.get(0);

        switch (winnerIndex)
        {
            case 0:
                return new Point((getWidth() - card.getWidth()) / 2, 556);
            case 1:
                return new Point(54, (getHeight() - card.getHeight()) / 2);
            case 2:
                return new Point((getWidth() - card.getWidth()) / 2, 18);
            case 3:
                return new Point(780, (getHeight() - card.getHeight()) / 2);
            default:
                throw new IllegalStateException();
        }
    }

    /**
     * Gets the list of human cards for change.
     */
    private List<CardPane> getHumansCardsForChange()
    {
        return cardsForChange.get(INDEX_OF_HUMAN);
    }

    /**
     * Gets the game which is currently displayed.
     */
    public Game getGame()
    {
        return game;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AudioClip getScreenMusic()
    {
        return AudioRepository.getInstance().getAudio("gameBackground.wav");
    }
}