package userInterface;

import gameElements.Game;
import gameElements.Player;
import repositories.AudioRepository;
import repositories.ImageRepository;
import sk.upjs.jpaz2.AudioClip;
import sk.upjs.jpaz2.ImageShape;
import sk.upjs.jpaz2.Turtle;
import userInterface.panes.Label;
import userInterface.panes.TwoStatedButton;
import userInterface.utilities.PaneUtilities;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;

/**
 * The intro screen of the game. The class is designed as a singleton.
 */
class IntroScreen extends Screen
{
    /**
     * The instance of the screen.
     */
    private static IntroScreen instance;

    /**
     * The font used for the buttons of the screen.
     */
    private static final Font BUTTON_FONT = new Font("Calibri", Font.CENTER_BASELINE, 28);

    /**
     * The color of the buttons' font.
     */
    private static final Color FONT_COLOR = Color.white;

    /**
     * The list of the all the clickable buttons on the screen.
     */
    private List<TwoStatedButton> buttons;

    /**
     * Constructs the intro screen.
     */
    private IntroScreen()
    {
        PaneUtilities.drawBackground(this, "introScreen.png");
        drawMusicIcon();
        drawLogo();
        initializeButtons();
    }

    /**
     * Draws the mandatory UPJS logo to the screen.
     */
    private void drawLogo()
    {
        Turtle logo = new Turtle();
        logo.setShape(ImageRepository.getInstance().getImageShape("logo.png"));
        logo.setPosition(466, 626);
        logo.setScale(0.2);
        add(logo);
    }

    /**
     * Initializes the screen buttons.
     */
    private void initializeButtons()
    {
        buttons = new ArrayList<>();
        ImageShape introButtonShape = ImageRepository.getInstance().getImageShape("introButton.png");
        String[] buttonNames = new String[]{"New game", "Statistics", "Rules", "Quit"};

        for (int i = 0; i < buttonNames.length; i++)
        {
            Label label = new Label(buttonNames[i], BUTTON_FONT, FONT_COLOR);
            TwoStatedButton button = new TwoStatedButton(introButtonShape, label);
            buttons.add(button);
            add(button);
            button.setPosition(getPositionOfButton(i));
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onMousePressed(int x, int y, MouseEvent detail)
    {
        TwoStatedButton button = getButton(x, y);

        if (button == null)
            return;

        playSound("click.wav");

        switch (buttons.indexOf(button))
        {
            case 0:
                newGame();
                break;
            case 1:
                Hearts.getInstance().changeScreen(StatisticsScreen.getInstance(), false);
                break;
            case 2:
                Hearts.getInstance().changeScreen(RulesScreen.getInstance(), false);
                break;
            case 3:
                System.exit(0);
                break;
        }

        button.turnOff();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onMouseMoved(int x, int y, MouseEvent detail)
    {
        TwoStatedButton button = getButton(x, y);
        buttons.forEach(TwoStatedButton::turnOff);

        if (button != null)
            button.turnOn();
    }

    /**
     * Creates a new game screen with a new game and binds it to the main window.
     */
    void newGame()
    {
        Hearts.getInstance().getCurrentScreen().turnOffMusic();

        List<Player> players = new ArrayList<>();

        // It's up to you to guess the meaning of these particular names.
        players.add(new Player("Player"));
        players.add(new Player("Frank"));
        players.add(new Player("Petr"));
        players.add(new Player("John"));

        Game newGame = new Game(players);
        GameScreen screen = new GameScreen(newGame);
        Hearts.getInstance().changeScreen(screen, true);
    }

    /**
     * Gets the button which contains the given coordinates.
     *
     * @param x The X-coordinate.
     * @param y The Y-coordinate.
     *
     * @return The button on the given coordinates, or null, if there is no such button.
     */
    private TwoStatedButton getButton(int x, int y)
    {
        TwoStatedButton result = null;

        for (TwoStatedButton button : buttons)
            if (button.containsPoint(x, y))
                result = button;

        return result;
    }

    /**
     * Gets the position of the button with the given index.
     */
    private Point getPositionOfButton(int index)
    {
        switch (index)
        {
            case 0:
                return new Point(80, 330);
            case 1:
                return new Point(578, 330);
            case 2:
                return new Point(80, 472);
            case 3:
                return new Point(578, 472);
            default:
                throw new IllegalStateException();
        }
    }

    /**
     * Gets the instance of the screen.
     */
    public static IntroScreen getInstance()
    {
        if (instance == null)
            instance = new IntroScreen();

        return instance;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AudioClip getScreenMusic()
    {
        return AudioRepository.getInstance().getAudio("introBackground.wav");
    }
}