package userInterface;

import gameElements.Game;
import gameElements.Player;
import repositories.ImageRepository;
import sk.upjs.jpaz2.AudioClip;
import sk.upjs.jpaz2.ImageShape;
import userInterface.panes.Button;
import userInterface.panes.Label;
import userInterface.panes.TwoStatedButton;
import userInterface.utilities.PaneUtilities;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

/**
 * A screen of a round result, displayed after each round.
 */
class RoundResultScreen extends Screen
{
    /**
     * The font of the screen.
     */
    private static final Font SCREEN_FONT = new Font("Calibri", Font.CENTER_BASELINE, 28);

    /**
     * The font color of the screen.
     */
    private static final Color FONT_COLOR = Color.white;

    /**
     * The given game screen which has displayed this screen.
     */
    private final GameScreen screen;

    /**
     * Indicates if the game, which's results the screen displays, has ended.
     */
    private final boolean gameEnded;

    /**
     * The list of labels of the first row of the screen table (the player's names).
     */
    private List<Label> firstRow;

    /**
     * The list of lists of labels representing the part of the screen table between the first and the last row
     * (exclusively).
     */
    private List<List<Label>> table;

    /**
     * The list of labels of the last row of the screen table (the player's total scores)
     */
    private List<Label> lastRow;

    /**
     * The continue button of the screen, or the new game button, if the game has ended.
     */
    private TwoStatedButton continueButton;

    /**
     * The main menu button of the screen.
     */
    private TwoStatedButton mainMenuButton;

    /**
     * Constructs a round results screen, which is going to be displayed by the given game screen.
     */
    RoundResultScreen(GameScreen screen)
    {
        this.screen = screen;
        gameEnded = getGame().hasGameEnded();
        PaneUtilities.drawBackground(this, "roundSummary.png");
        initializeLabels();
        drawTable();
        drawButtons();
        drawMusicIcon();
    }

    /**
     * Initialize labels of the table of the screen.
     */
    private void initializeLabels()
    {
        firstRow = new ArrayList<>();
        firstRow.add(new Label("Round", SCREEN_FONT, FONT_COLOR));

        // player's names
        firstRow.addAll(getGame().getPlayers().stream()
                .map(player -> new Label(player.toString(), SCREEN_FONT, FONT_COLOR))
                .collect(Collectors.toList()));

        table = new ArrayList<>();

        for (int i = 1; i <= getGame().getCurrentRoundNumber(); i++)
        {
            List<Label> row = new ArrayList<>();

            row.add(new Label(Integer.toString(i), SCREEN_FONT, FONT_COLOR));

            // player's scores in the i-th round
            for (Player player : getGame().getPlayers())
                row.add(new Label(Integer.toString(player.getScoreOfRound(i)), SCREEN_FONT, FONT_COLOR));

            table.add(row);
        }

        lastRow = new ArrayList<>();
        lastRow.add(new Label("Overall", SCREEN_FONT, FONT_COLOR));

        // overall player's scores
        lastRow.addAll(getGame().getPlayers().stream()
                .map(player -> new Label(Integer.toString(player.getCurrentGameScore()), SCREEN_FONT, FONT_COLOR))
                .collect(Collectors.toList()));
    }

    /**
     * Draw the screen table to the screen.
     */
    private void drawTable()
    {
        Label round = firstRow.get(0);
        add(round);
        round.setPosition(57 + (105 - round.getWidth()) / 2, 122 + (47 - round.getHeight()) / 2);

        for (int i = 1; i < firstRow.size(); i++)
        {
            Label name = firstRow.get(i);
            add(name);
            name.setPosition(173 * i + (163 - name.getWidth()) / 2, 122 + (47 - name.getHeight()) / 2);
        }

        // We can't display more than 7 rounds according to the fixed table height. So if there are more
        // than 7 rounds, we'll display the last 7
        final int rowsToBeDisplayed = Math.min(table.size(), 7);
        final int firstRowToBeDisplayed = Math.max(0, table.size() - 7);

        for (int i = 0; i < rowsToBeDisplayed; i++)
        {
            Label number = table.get(firstRowToBeDisplayed + i).get(0);
            add(number);
            number.setPosition(57 + (105 - number.getWidth()) / 2, 178 + i * (number.getHeight() + 10));

            for (int j = 1; j <= Game.NUMBER_OF_PLAYERS; j++)
            {
                Label score = table.get(firstRowToBeDisplayed + i).get(j);
                add(score);
                score.setPosition(173 * j + (163 - score.getWidth()) / 2, 178 + i * (score.getHeight() + 10));
            }
        }

        Label overall = lastRow.get(0);
        add(overall);
        overall.setPosition(57 + (105 - overall.getWidth()) / 2, 446 + (47 - overall.getHeight()) / 2);

        for (int i = 1; i < firstRow.size(); i++)
        {
            Label score = lastRow.get(i);
            add(score);
            score.setPosition(173 * i + (163 - score.getWidth()) / 2, 446 + (47 - score.getHeight()) / 2);
        }
    }

    /**
     * Draws the buttons to the screen.
     */
    private void drawButtons()
    {
        ImageShape button = ImageRepository.getInstance().getImageShape("roundSummaryButton.png");

        Label intro = new Label("Main menu", SCREEN_FONT, FONT_COLOR);
        mainMenuButton = new TwoStatedButton(button, intro);
        add(mainMenuButton);
        mainMenuButton.setPosition(498, 507);

        Label nextRound = new Label(gameEnded ? "New game" : "Continue", SCREEN_FONT, FONT_COLOR);
        continueButton = new TwoStatedButton(button, nextRound);
        add(continueButton);
        continueButton.setPosition(191, 507);

        initializeEndGameButton();
    }

    /**
     * Initialize the end game button. If the game hasn't ended yet, it does nothing.
     */
    private void initializeEndGameButton()
    {
        if (!gameEnded)
            return;

        ImageShape winOrLose = ImageRepository.getInstance().getImageShape("endGameButton.png");
        Label endGameButtonLabel = getEndGameButtonLabel();
        Button middleButton = new Button(winOrLose, endGameButtonLabel);
        add(middleButton);
        middleButton.setPosition(342, 276);
    }

    /**
     * Gets the end game label which will be displayed in the button.
     */
    private Label getEndGameButtonLabel()
    {
        switch (getGame().getIndexOfWinner())
        {
            case GameScreen.INDEX_OF_HUMAN:
                return new Label("You win!", SCREEN_FONT, FONT_COLOR);
            case -1:
                return new Label("Draw!", SCREEN_FONT, FONT_COLOR);
            default:
                return new Label("You lose!", SCREEN_FONT, FONT_COLOR);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onMousePressed(int x, int y, MouseEvent detail)
    {
        if (continueButton.containsPoint(x, y))
        {
            playSound("click.wav");

            if (gameEnded)
                IntroScreen.getInstance().newGame();
            else
                screen.continueInGame();

            continueButton.turnOff();
        }

        if (mainMenuButton.containsPoint(x, y))
        {
            playSound("click.wav");
            screen.turnOffMusic();
            Hearts.getInstance().changeScreen(IntroScreen.getInstance(), true);
            mainMenuButton.turnOff();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onMouseMoved(int x, int y, MouseEvent detail)
    {
        TwoStatedButton button = null;

        if (continueButton.containsPoint(x, y))
            button = continueButton;

        if (mainMenuButton.containsPoint(x, y))
            button = mainMenuButton;

        continueButton.turnOff();
        mainMenuButton.turnOff();

        if (button != null)
            button.turnOn();
    }

    /**
     * Gets the game which's result the screen displays.
     */
    private Game getGame()
    {
        return screen.getGame();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AudioClip getScreenMusic()
    {
        return screen.getScreenMusic();
    }
}