package userInterface;

import repositories.GameRepository;
import repositories.ImageRepository;
import sk.upjs.jpaz2.AudioClip;
import sk.upjs.jpaz2.ImageShape;
import userInterface.panes.Label;
import userInterface.panes.TwoStatedButton;
import userInterface.utilities.PaneUtilities;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;

/**
 * Represents a screen with the statistics of all games. It is designed as a singleton.
 */
class StatisticsScreen extends Screen
{
    /**
     * The instance of the screen.
     */
    private static StatisticsScreen instance;

    /**
     * The font used for the buttons of the screen.
     */
    private static final Font BUTTON_FONT = new Font("Calibri", Font.CENTER_BASELINE, 28);

    /**
     * The font used for the table of the screen.
     */
    private static final Font TABLE_FONT = BUTTON_FONT.deriveFont(25f);

    /**
     * The color of the screen font.
     */
    private static final Color FONT_COLOR = Color.white;

    /**
     * The reset stats button of the screen.
     */
    private TwoStatedButton resetStats;

    /**
     * The main menu button of the screen.
     */
    private TwoStatedButton mainMenu;

    /**
     * The list of labels which display the stats.
     */
    private List<Label> stats;

    /**
     * Constructs the statistics screen.
     */
    private StatisticsScreen()
    {
        PaneUtilities.drawBackground(this, "statisticsScreen.png");
        initializeButtons();
        drawLegend();
        drawStats();
        drawMusicIcon();
    }

    /**
     * Draws the legend to the screen.
     */
    private void drawLegend()
    {
        List<Label> legend = new ArrayList<>();

        legend.add(new Label("Games Played", TABLE_FONT, FONT_COLOR));
        legend.add(new Label("Games Won", TABLE_FONT, FONT_COLOR));
        legend.add(new Label("Winning Percentage", TABLE_FONT, FONT_COLOR));
        legend.add(new Label("Total Rounds", TABLE_FONT, FONT_COLOR));
        legend.add(new Label("Avg. Points per Round", TABLE_FONT, FONT_COLOR));
        legend.add(new Label("Moon Shoots", TABLE_FONT, FONT_COLOR));

        for (int i = 0; i < legend.size(); i++)
        {
            Label current = legend.get(i);
            add(current);
            current.setPosition(275, 211 + 49 * i);
        }
    }

    /**
     * Initializes the screen buttons.
     */
    private void initializeButtons()
    {
        ImageShape button = ImageRepository.getInstance().getImageShape("statisticsScreenButtons.png");

        Label resetLabel = new Label("Reset stats", BUTTON_FONT, FONT_COLOR);
        resetStats = new TwoStatedButton(button, resetLabel);
        add(resetStats);
        resetStats.setPosition(243, 512);

        Label mainLabel = new Label("Main menu", BUTTON_FONT, FONT_COLOR);
        mainMenu = new TwoStatedButton(button, mainLabel);
        add(mainMenu);
        mainMenu.setPosition(465, 512);
    }

    /**
     * Loads and draws the current statistics to the screen table.
     */
    private void drawStats()
    {
        if (stats == null)
            stats = new ArrayList<>();

        // We'll reset stats such that we remove all the stat labels from the screen and reset the stats list
        stats.forEach(this::remove);
        stats.clear();

        List<String> stringStats = GameRepository.getInstance().getCurrentStatsAsListOfStrings();

        for (int i = 0; i < stringStats.size(); i++)
        {
            Label current = new Label(stringStats.get(i), TABLE_FONT, FONT_COLOR);
            stats.add(current);
            add(current);
            current.setPosition(643 - current.getWidth(), 211 + 49 * i);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onMousePressed(int x, int y, MouseEvent detail)
    {
        if (resetStats.containsPoint(x, y))
        {
            playSound("click.wav");
            GameRepository.getInstance().resetStats();
            drawStats();
            resetStats.turnOff();
            return;
        }

        if (mainMenu.containsPoint(x, y))
        {
            playSound("click.wav");
            Hearts.getInstance().changeScreen(IntroScreen.getInstance(), false);
            mainMenu.turnOff();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onMouseMoved(int x, int y, MouseEvent detail)
    {
        TwoStatedButton button = null;

        if (resetStats.containsPoint(x, y))
            button = resetStats;

        if (mainMenu.containsPoint(x, y))
            button = mainMenu;

        resetStats.turnOff();
        mainMenu.turnOff();

        if (button != null)
            button.turnOn();
    }

    /**
     * Gets the instance of the screen.
     */
    public static StatisticsScreen getInstance()
    {
        if (instance == null)
            instance = new StatisticsScreen();

        return instance;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AudioClip getScreenMusic()
    {
        return IntroScreen.getInstance().getScreenMusic();
    }
}