package userInterface.panes;

import gameElements.Card;
import repositories.ImageRepository;
import userInterface.utilities.PaneUtilities;

/**
 * A pane, which represents a card, either with a value, or from the back.
 */
public class CardPane extends PaneWithShadow
{
    /**
     * The card which is displayed by the pane, or null, if this pane represents a card from the back.
     */
    private Card card;

    /**
     * Indicates whether the card is raised.
     */
    private boolean raised;

    /**
     * Indicates whether the card can be raised.
     */
    private boolean canBeRaised = true;

    /**
     * The height of raise of the raised card, or 0 if the card is not raised.
     */
    private int heightOfRaise;

    /**
     * Constructs a card pane which displays a given card.
     */
    public CardPane(Card card)
    {
        this.card = card;
        drawBackground();
    }

    /**
     * Constructs a card pane which displays a card from the back.
     */
    public CardPane()
    {
        this(null);
    }

    /**
     * Draws the background to the card.
     */
    private void drawBackground()
    {
        setBorderWidth(0);
        setTransparentBackground(true);
        PaneUtilities.drawBackground(this, ImageRepository.getInstance().getImageShape(card));
    }

    /**
     * Raises the card by the given height. To raise card means to increase it's Y-coordinate. If it's been already
     * raised, it won't do anything.
     */
    public void raise(int height)
    {
        if (raised)
            return;

        // We change the pane height because if we don't do that and our cursor is
        // in the bottom of the card and there is another card bellow this that can be clicked on,
        // then the card bellow could react on events, which wouldn't look well if one of the events
        // is to raise it (which is in the OnMouseMoved method of a GameScreen).
        setHeight(getHeight() + height);
        setY(getY() - height);
        heightOfRaise = height;
        raised = true;
    }

    /**
     * Lower the card if it was raised. Otherwise does nothing. A raised card is a card whose Y-coordinate has been
     * increased by some height.
     */
    public void lower()
    {
        if (!raised)
            return;

        // for the reason why we've changed the height of the pane, see the internal
        // commentary in the raise(int height) method of this class
        setHeight(getHeight() - heightOfRaise);
        setY(getY() + heightOfRaise);
        heightOfRaise = 0;
        raised = false;
    }

    /**
     * Gets the card which is displayed by the pane. If the pane displays a card from the back, it returns null.
     */
    public Card getCard()
    {
        return card;
    }

    /**
     * Sets the card which is displayed by the pane. If we want to display a card from the back, we must set the card to
     * be null.
     */
    public void setCard(Card card)
    {
        this.card = card;
        drawBackground();
    }

    /**
     * Gets if the card is raised. A raised card is a card which's Y-coordinate has been increased by some height.
     */
    public boolean isRaised()
    {
        return raised;
    }

    /**
     * Gets if the pane can be raised. It's meant that a card can be raised if and only if you can click on it and play
     * it or change it.
     */
    public boolean isCanBeRaised()
    {
        return canBeRaised;
    }

    /**
     * Sets if the pane can be raised. It's meant that a card can be raised if and only if you can click on it and play
     * it or change it.
     */
    public void setCanBeRaised(boolean canBeRaised)
    {
        this.canBeRaised = canBeRaised;
    }
}