package userInterface.panes;

import sk.upjs.jpaz2.ImageShape;
import sk.upjs.jpaz2.Turtle;

/**
 * Represents a two-stated button. It's meant to be used for a button which you can click on. It would be in the on
 * state if the mouse cursor is over the button, otherwise it would be in the off state. Although, this functionality
 * must be implemented in a screen class, because there is no good way to find out whether the current cursor position
 * is outside the pane (i.e. outside the button) so we can't simply turn it off.
 */
public class TwoStatedButton extends Button
{
    /**
     * The turtle which graphically represents the button. It has both views inside.
     */
    private Turtle button;

    /**
     * Indicates whether the button reacts if the cursor is over it.
     */
    private boolean reactive = true;

    /**
     * Constructs a two-stated button with the given background and the given label. It is assumed that the given shape
     * has at least two views, such that the first one is the on state.
     */
    public TwoStatedButton(ImageShape shape, Label label)
    {
        super(shape, label);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void drawBackground()
    {
        setBorderWidth(0);
        setTransparentBackground(true);
        button = new Turtle();
        button.setShape(background);
        // the default state would be off
        button.setViewIndex(1);
        add(button);
        button.center();
    }

    /**
     * Displays the clicked state.
     */
    public void turnOn()
    {
        button.setViewIndex(0);
    }

    /**
     * Displays the non-clicked state.
     */
    public void turnOff()
    {
        button.setViewIndex(1);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected boolean onCanClick(int x, int y)
    {
        return reactive;
    }

    /**
     * Sets if the button is reactive, which means whether the button reacts if the cursor is over it. If the cursor is
     * over the button, then the hand cursor will be displayed if and only if this is true.
     */
    public void setReactive(boolean reactive)
    {
        this.reactive = reactive;
    }

    /**
     * Gets if the button is reactive, which means whether the button reacts if the cursor is over it. If the cursor is
     * over the button, then the hand cursor will be displayed if and only if this is true.
     */
    public boolean isReactive()
    {
        return reactive;
    }
}