package userInterface.utilities;

import sk.upjs.jpaz2.TickTimer;

import java.util.function.Consumer;

/**
 * Provides static methods for timing internally using {@link TickTimer}.
 */
public class TimerUtilities
{
    /**
     * Private constructor forbidding creation of class instances.
     */
    private TimerUtilities()
    {
    }

    /**
     * A {@link TickTimer} with the ability to set it's periodical function, which is a function of an index of a
     * current tick.
     */
    public static class Timer extends TickTimer
    {
        /**
         * The consumer of the timer, which as the argument takes the index of the current tick.
         */
        private Consumer<Integer> consumer;

        /**
         * The current tick index.
         */
        int currentTick;

        /**
         * Sets the consumer, which is a function of a current tick index. It might be null, then it's interpreted as an
         * empty function.
         */
        public void setConsumer(Consumer<Integer> consumer)
        {
            this.consumer = consumer;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        protected void onTick()
        {
            if (consumer != null)
                consumer.accept(currentTick);

            currentTick++;
        }
    }

    /**
     * Gets a periodical timer with the given properties.
     *
     * @param period        The period (in milliseconds).
     * @param numberOfTicks The number of ticks.
     * @param consumer      The function of the current tick index.
     * @param actionAfter   The action performed after all ticks.
     */
    private static Timer getPeriodicalTimer(int period, int numberOfTicks, Consumer<Integer> consumer, Runnable actionAfter)
    {
        Timer t = new Timer()
        {
            /**
             * {@inheritDoc}
             */
            @Override
            protected void onTick()
            {
                super.onTick();

                if (currentTick == numberOfTicks)
                {
                    setTickPeriod(0);

                    if (actionAfter != null)
                        actionAfter.run();
                }
            }

            /**
             * {@inheritDoc}
             */
            @Override
            public void setEnabled(boolean enabled)
            {
                setTickPeriod(period);
                super.setEnabled(enabled);
            }
        };

        t.setConsumer(consumer);

        return t;
    }

    /**
     * Gets a timer with the given period (in milliseconds), the given number of ticks, with no periodical action. It
     * might be set later.
     */
    public static Timer getPlainPeriodicalTimer(int period, int numberOfTicks)
    {
        return getPeriodicalTimer(period, numberOfTicks, null, null);
    }

    /**
     * Waits a given amount of time (in milliseconds), for a given action. It internally uses {@link Timer}, which
     * creates a new thread. Therefore, if you invoke for example waitFor(100,action1), waitFor(200,action2), in a row,
     * the second action won't be invoked after 300 millisecond. If you want exactly that, you have to nest the calls
     * (it means to make sure that the second waitFor is called from inside the action1).
     */
    public static void waitFor(int timeToWait, Runnable action)
    {
        getPeriodicalTimer(timeToWait, 1, null, action).setEnabled(true);
    }

    /**
     * Executes the given periodical action (which is a function of an index of a current tick), with the given period,
     * for the given number of ticks, with the given action after the execution.
     */
    public static void executePeriodically(int period, int numberOfTicks, Consumer<Integer> periodicalAction, Runnable actionAfter)
    {
        getPeriodicalTimer(period, numberOfTicks, periodicalAction, actionAfter).setEnabled(true);
    }
}